from __future__ import annotations

import argparse
import json
import os
import sys

from . import config, constants, printjson, update
from . import messages as m


def main() -> None:
    # Hack around argparse's lack of optional subparsers
    if len(sys.argv) == 1:
        sys.argv.append("spec")

    try:
        with open(config.scriptPath("semver.txt"), encoding="utf-8") as fh:
            semver = fh.read().strip()
            semverText = f"Bikeshed v{semver}: "
    except FileNotFoundError:
        semver = "???"
        semverText = ""

    argparser = argparse.ArgumentParser(description=f"{semverText}Processes spec source files into valid HTML.")
    argparser.add_argument("--version", action="version", version=semver)
    argparser.add_argument(
        "-q",
        "--quiet",
        dest="quiet",
        action="count",
        default=0,
        help="Silences one level of message, least-important first.",
    )
    argparser.add_argument(
        "-s",
        "--silent",
        dest="silent",
        action="store_true",
        help="Shorthand for 'as many -q as you need to shut it up'",
    )
    argparser.add_argument(
        "-f",
        "--force",
        dest="errorLevel",
        action="store_const",
        const="nothing",
        help="Force the preprocessor to run to completion; fatal errors don't stop processing.",
    )
    argparser.add_argument(
        "-d",
        "--dry-run",
        dest="dryRun",
        action="store_true",
        help="Prevents the processor from actually saving anything to disk, but otherwise fully runs.",
    )
    argparser.add_argument(
        "-a",
        "--ascii-only",
        dest="asciiOnly",
        action="store_true",
        help="Force all Bikeshed messages to be ASCII-only.",
    )
    argparser.add_argument(
        "--print",
        dest="printMode",
        choices=m.PRINT_MODES,
        default=None,
        help="How Bikeshed formats its message output. Options are 'plain' (just text), 'console' (text with console color codes), 'markup' (XML), and 'json' (JSON stream). Defaults to 'console'.",
    )
    argparser.add_argument(
        "--die-on",
        dest="errorLevel",
        choices=list(m.MESSAGE_LEVELS.keys()),
        help="Determines what sorts of errors cause Bikeshed to die (refuse to generate an output document). Default is 'fatal'; the -f flag is a shorthand for 'nothing'",
    )
    argparser.add_argument(
        "--die-when",
        dest="errorTiming",
        choices=m.DEATH_TIMING,
        default="late",
        help="When a disallowed error should force Bikeshed to stop. 'early' causes it to stop immediately so you can deal with the first error; 'late' makes it process the entire document first and only stop at the end so you can see all the errors.",
    )
    argparser.add_argument(
        "--no-update",
        dest="skipUpdate",
        action="store_true",
        help="Skips checking if your data files are up-to-date.",
    )
    argparser.add_argument(
        "--allow-nonlocal-files",
        dest="allowNonlocalFiles",
        action="store_true",
        help="Allows Bikeshed to see/include files from folders higher than the one your source document is in.",
    )
    argparser.add_argument(
        "--allow-execute",
        dest="allowExecute",
        action="store_true",
        help="Allow some features to execute arbitrary code from outside the Bikeshed codebase.",
    )

    subparsers = argparser.add_subparsers(title="Subcommands", dest="subparserName")

    specParser = subparsers.add_parser("spec", help="Process a spec source file into a valid output file.")
    specParser.add_argument(
        "infile",
        nargs="?",
        default=None,
        help='Path to the source file: stdin ("-"), an https URL, a .bs or.src.html file, or a tar archive containing an index.bs file.',
    )
    specParser.add_argument(
        "outfile",
        nargs="?",
        default=None,
        help='Path to the output file: stdout ("-"), or a filename.',
    )
    specParser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        help="Switches on some debugging tools. Don't use for production!",
    )
    specParser.add_argument(
        "--gh-token",
        dest="ghToken",
        nargs="?",
        help="GitHub access token. Useful to avoid API rate limits. Generate tokens: https://github.com/settings/tokens.",
    )
    specParser.add_argument(
        "--byos",
        dest="byos",
        action="store_true",
        help="Bring-Your-Own-Spec: turns off all the Bikeshed auto-niceties, so you can piecemeal its features into your existing doc instead. Experimental, let me know if things get crashy or weird.",
    )
    specParser.add_argument(
        "-l",
        "--line-numbers",
        dest="lineNumbers",
        action="store_true",
        help="Hacky support for outputting line numbers on all error messages. Disables output, as this is hacky and might mess up your source.",
    )

    echidnaParser = subparsers.add_parser(
        "echidna",
        help="Process a spec source file into a valid output file and publish it according to certain automatic protocols.",
    )
    echidnaParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    echidnaParser.add_argument("--u", dest="un", metavar="USERNAME", required=False, help="W3C username.")
    echidnaParser.add_argument("--p", dest="pw", metavar="PASSWORD", required=False, help="W3C password.")
    echidnaParser.add_argument(
        "--decision",
        dest="decision",
        metavar="URL",
        required=False,
        help="URL recording the decision to publish.",
    )
    echidnaParser.add_argument(
        "--gh-token",
        dest="ghToken",
        required=False,
        help="GitHub access token. Useful to avoid API rate limits. Generate tokens: https://github.com/settings/tokens.",
    )
    echidnaParser.add_argument(
        "--timeout",
        dest="timeout",
        metavar="SECONDS",
        required=False,
        default=3,
        type=float,
        help="The timeout for contacting the publication service (default 3s).",
    )
    echidnaParser.add_argument(
        "--editorial",
        dest="editorial",
        action="store_true",
        required=False,
        help="Flags update as editorial.",
    )
    echidnaParser.add_argument(
        "--cc",
        dest="cc",
        metavar="EMAIL",
        required=False,
        help="Comma-separated list of email addresses to ping with the publication status when complete.",
    )
    echidnaParser.add_argument(
        "--additional-directories",
        dest="additionalDirectories",
        required=False,
        nargs="*",
        help="Directories to bundle in the tar file. Defaults to examples/, diagrams/, and images/.",
    )
    echidnaParser.add_argument(
        "--self-contained",
        dest="selfContained",
        action="store_true",
        help="The spec is self-contained, do not bundle any extra directories in the tar file.",
    )
    echidnaParser.add_argument(
        "--just-tar",
        dest="justTar",
        action="store_true",
        help="Don't actually submit to the echidna service, instead just echo the prepared TAR file to stdout.",
    )

    watchParser = subparsers.add_parser(
        "watch",
        help="Process a spec source file into a valid output file, automatically rebuilding when it changes.",
    )
    watchParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    watchParser.add_argument("outfile", nargs="?", default=None, help="Path to the output file.")
    watchParser.add_argument(
        "--gh-token",
        dest="ghToken",
        nargs="?",
        help="GitHub access token. Useful to avoid API rate limits. Generate tokens: https://github.com/settings/tokens.",
    )
    watchParser.add_argument(
        "--byos",
        dest="byos",
        action="store_true",
        help="Bring-Your-Own-Spec: turns off all the Bikeshed auto-niceties, so you can piecemeal its features into your existing doc instead. Experimental, let me know if things get crashy or weird.",
    )

    serveParser = subparsers.add_parser("serve", help="Identical to 'watch', but also serves the folder on localhost.")
    serveParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    serveParser.add_argument("outfile", nargs="?", default=None, help="Path to the output file.")
    serveParser.add_argument(
        "--port",
        dest="port",
        nargs="?",
        default="8000",
        help="Specify the port to serve it over.",
    )
    serveParser.add_argument(
        "--localhost",
        dest="localhost",
        action="store_true",
        help="Only allow connections from localhost.",
    )
    serveParser.add_argument(
        "--gh-token",
        dest="ghToken",
        nargs="?",
        help="GitHub access token. Useful to avoid API rate limits. Generate tokens: https://github.com/settings/tokens.",
    )
    serveParser.add_argument(
        "--byos",
        dest="byos",
        action="store_true",
        help="Bring-Your-Own-Spec: turns off all the Bikeshed auto-niceties, so you can piecemeal its features into your existing doc instead. Experimental, let me know if things get crashy or weird.",
    )

    updateParser = subparsers.add_parser(
        "update",
        help="Update supporting files (those in /spec-data).",
        epilog="If no options are specified, everything is downloaded.",
    )
    updateModeGroup = updateParser.add_mutually_exclusive_group()
    updateModeGroup.add_argument(
        "--skip-manifest",
        dest="updateMode",
        action="store_const",
        const=update.UpdateMode.MANUAL,
        help="Forces Bikeshed to do a full update manually, rather than using the manifest to get the preprocessed update (which can be several minutes old).",
    )
    updateModeGroup.add_argument(
        "--force-manifest",
        dest="updateMode",
        action="store_const",
        const=update.UpdateMode.MANIFEST | update.UpdateMode.FORCE,
        help="Force a manifest-based update even if local manifest data is more recent than the remote manifest.",
    )
    updateModeGroup.set_defaults(
        updateMode=update.UpdateMode.BOTH,
    )
    updateParser.add_argument("--anchors", action="store_true", help="Download crossref anchor data.")
    updateParser.add_argument("--backrefs", action="store_true", help="Download link backref data.")
    updateParser.add_argument("--biblio", action="store_true", help="Download biblio data.")
    updateParser.add_argument("--boilerplate", action="store_true", help="Download boilerplate files.")
    updateParser.add_argument("--caniuse", action="store_true", help="Download Can I Use... data.")
    updateParser.add_argument("--mdn", action="store_true", help="Download MDN Spec Links... data.")
    updateParser.add_argument(
        "--link-defaults",
        dest="linkDefaults",
        action="store_true",
        help="Download link default data.",
    )
    updateParser.add_argument(
        "--languages",
        dest="languages",
        action="store_true",
        help="Download language/translation data.",
    )
    updateParser.add_argument(
        "--wpt",
        dest="wpt",
        action="store_true",
        help="Download web-platform-tests data.",
    )

    issueParser = subparsers.add_parser(
        "issues-list",
        help="Process a plain-text issues file into HTML. Call with no args to see an example input text.",
    )
    issueParser.add_argument(
        "-t",
        dest="printTemplate",
        action="store_true",
        help="Output example Issues List template.",
    )
    issueParser.add_argument("infile", nargs="?", default=None, help="Path to the plain-text issue file.")
    issueParser.add_argument(
        "outfile",
        nargs="?",
        default=None,
        help="Path to the output file. Default is file of the same name as input, with .html.",
    )

    debugParser = subparsers.add_parser("debug", help="Run various debugging commands.")
    debugParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    debugCommands = debugParser.add_mutually_exclusive_group(required=True)
    debugCommands.add_argument(
        "--print-exports",
        dest="printExports",
        action="store_true",
        help="Prints those terms that will be exported for cross-ref purposes.",
    )
    debugCommands.add_argument("--print", dest="code", help="Runs the specified code and prints it.")
    debugCommands.add_argument(
        "--print-json",
        dest="jsonCode",
        help="Runs the specified code and prints it as formatted JSON.",
    )
    debugCommands.add_argument(
        "--refresh-data",
        dest="refreshData",
        action="store_true",
        help="Clobbers the readonly data files with the mutable ones.",
    )
    debugCommands.add_argument(
        "--print-metadata",
        dest="printMetadata",
        action="store_true",
        help="Prints all the metadata parsed for the spec as JSON. Top-level keys are presented in increasing order of importance; second-level keys are in order of first appearance in each context.",
    )

    refParser = subparsers.add_parser("refs", help="Search Bikeshed's ref database.")
    refParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    refParser.add_argument("--text", dest="text", default=None)
    refParser.add_argument("--type", dest="linkType", default=None)
    refParser.add_argument("--for", dest="linkFor", default=None)
    refParser.add_argument("--spec", dest="spec", default=None)
    refParser.add_argument("--status", dest="status", default=None)
    refParser.add_argument(
        "--exact",
        dest="exact",
        action="store_true",
        help="Only search for the exact text provided; don't apply Bikeshed's automatic conjugation help for plurals/etc.",
    )
    refParser.add_argument(
        "--latest-only",
        dest="latestOnly",
        action="store_true",
        help="Apply Bikeshed's logic for only returning the latest version of a given ref when it exists in multiple levels of a spec.",
    )

    sourceParser = subparsers.add_parser("source", help="Tools for formatting the *source* document.")
    sourceParser.add_argument(
        "--big-text",
        dest="bigText",
        action="store_true",
        help="Finds HTML comments containing 'Big Text: foo' and turns them into comments containing 'foo' in big text.",
    )
    sourceParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")
    sourceParser.add_argument("outfile", nargs="?", default=None, help="Path to the output file.")

    outlineParser = subparsers.add_parser("outline", help="Generates an outline for the spec.")
    outlineParser.add_argument("infile", nargs="?", default=None, help="Path to the source file.")

    testParser = subparsers.add_parser("test", help="Tools for running Bikeshed's testsuite.")
    testParser.add_argument(
        "--rebase",
        default=False,
        action="store_true",
        help="Rebase the specified files.",
    )
    testParser.add_argument(
        "--manual",
        dest="manualOnly",
        default=False,
        action="store_true",
        help="Skip testing the real-world files in the repo, and only run the manually-written ones.",
    )
    testParser.add_argument(
        "--folder",
        dest="folders",
        default=None,
        nargs="+",
        help="Only run tests whose paths contain any of these folder names.",
    )
    testParser.add_argument(
        "--file",
        dest="files",
        default=None,
        nargs="+",
        help="Only run tests whose filenames contain any of these strings as substrings.",
    )

    profileParser = subparsers.add_parser(
        "profile",
        help="Profiling Bikeshed. Needs graphviz, gprof2dot, and xdot installed.",
    )
    profileParser.add_argument(
        "--root",
        dest="root",
        default=None,
        metavar="ROOTFUNC",
        help="Prune the graph to start with the specified root node.",
    )
    profileParser.add_argument(
        "--leaf",
        dest="leaf",
        default=None,
        metavar="LEAFFUNC",
        help="Prune the graph to only show ancestors of the specified leaf node.",
    )
    profileParser.add_argument(
        "--svg",
        dest="svgFile",
        default=None,
        help="Save the graph to a specified SVG file, rather than outputting with xdot immediately.",
    )

    templateParser = subparsers.add_parser("template", help="Outputs a skeleton .bs file for you to start with.")
    templateParser.add_argument(
        "variant",
        choices=["base", "minimal", "test"],
        nargs="?",
        default="base",
    )

    wptParser = subparsers.add_parser("wpt", help="Tools for writing Web Platform Tests.")
    wptParser.add_argument(
        "--template",
        default=False,
        action="store_true",
        help="Outputs a skeleton WPT file for you to start with.",
    )

    options, extras = argparser.parse_known_args()

    if options.silent:
        m.state.printOn = "nothing"
        m.state.silent = True
    else:
        m.state.printOn = m.MessagesState.categoryName(options.quiet)
    if options.errorLevel is not None:
        m.state.dieOn = options.errorLevel
    m.state.dieWhen = options.errorTiming
    m.state.asciiOnly = options.asciiOnly
    if options.printMode is None:
        if "NO_COLOR" in os.environ or os.environ.get("TERM") == "dumb":
            m.state.printMode = "plain"
        else:
            m.state.printMode = "console"
    else:
        m.state.printMode = options.printMode
    constants.dryRun = options.dryRun
    constants.chroot = not options.allowNonlocalFiles
    constants.executeCode = options.allowExecute

    m.printOpener()

    if options.subparserName in ("spec", "echnida", "watch", "serve", "refs"):
        updateMode = update.UpdateMode.NONE if options.skipUpdate else update.UpdateMode.BOTH
        update.fixupDataFiles(updateMode=updateMode)
    if options.subparserName == "update":
        handleUpdate(options)
    elif options.subparserName == "spec":
        handleSpec(options, extras)
    elif options.subparserName == "echidna":
        handleEchidna(options, extras)
    elif options.subparserName == "watch":
        handleWatch(options, extras)
    elif options.subparserName == "serve":
        handleServe(options, extras)
    elif options.subparserName == "debug":
        handleDebug(options, extras)
    elif options.subparserName == "refs":
        handleRefs(options, extras)
    elif options.subparserName == "issues-list":
        handleIssuesList(options)
    elif options.subparserName == "source":
        handleSource(options)
    elif options.subparserName == "outline":
        handleOutline(options)
    elif options.subparserName == "test":
        handleTest(options, extras)
    elif options.subparserName == "profile":
        handleProfile(options)
    elif options.subparserName == "template":
        handleTemplate(options)
    elif options.subparserName == "wpt":
        handleWpt(options)

    m.printCloser()


def handleUpdate(options: argparse.Namespace) -> None:
    update.update(
        anchors=options.anchors,
        backrefs=options.backrefs,
        biblio=options.biblio,
        boilerplate=options.boilerplate,
        caniuse=options.caniuse,
        mdn=options.mdn,
        linkDefaults=options.linkDefaults,
        languages=options.languages,
        wpt=options.wpt,
        dryRun=constants.dryRun,
        updateMode=options.updateMode,
    )


def handleSpec(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata
    from .Spec import Spec

    doc = Spec(
        inputFilename=options.infile,
        debug=options.debug,
        token=options.ghToken,
        lineNumbers=options.lineNumbers,
    )
    if not doc.valid:
        m.die("Spec is in an invalid state; exitting.")
        return
    doc.mdCommandLine = metadata.fromCommandLine(extras)
    if options.errorLevel:
        doc.mdCommandLine.addData("Die On", options.errorLevel)
    if options.errorTiming:
        doc.mdCommandLine.addData("Die When", options.errorTiming)
    if options.byos:
        doc.mdCommandLine.addData("Group", "byos")
    doc.preprocess()
    doc.finish(outputFilename=options.outfile)


def handleEchidna(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata, publish
    from .Spec import Spec

    doc = Spec(inputFilename=options.infile, token=options.ghToken)
    if not doc.valid:
        m.die("Spec is in an invalid state; exitting.")
        return
    doc.mdCommandLine = metadata.fromCommandLine(extras)
    doc.mdCommandLine.addData("Prepare For TR", "yes")
    doc.preprocess()
    addDirs = [] if options.selfContained else options.additionalDirectories
    if options.justTar:
        tarBytes = publish.prepareTar(doc, additionalDirectories=addDirs)
        sys.stdout.buffer.write(tarBytes)
    else:
        publish.publishEchidna(
            doc,
            username=options.un,
            password=options.pw,
            decision=options.decision,
            additionalDirectories=addDirs,
            cc=options.cc,
            editorial=options.editorial,
            timeout=options.timeout,
        )


def handleWatch(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata
    from .Spec import Spec

    doc = Spec(inputFilename=options.infile, token=options.ghToken)
    if not doc.valid:
        m.die("Spec is in an invalid state; exitting.")
        return
    doc.mdCommandLine = metadata.fromCommandLine(extras)
    if options.byos:
        doc.mdCommandLine.addData("Group", "byos")
    doc.watch(outputFilename=options.outfile)


def handleServe(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata
    from .Spec import Spec

    doc = Spec(inputFilename=options.infile, token=options.ghToken)
    if not doc.valid:
        m.die("Spec is in an invalid state; exitting.")
        return
    doc.mdCommandLine = metadata.fromCommandLine(extras)
    # Can't have an error killing the watcher
    doc.mdCommandLine.addData("Die On", "nothing")
    if options.byos:
        doc.mdCommandLine.addData("Group", "byos")
    doc.watch(outputFilename=options.outfile, port=int(options.port))


def handleDebug(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata
    from .Spec import Spec

    m.state.dieOn = "nothing"
    m.state.printOn = "fatal"
    if options.printExports:
        doc = Spec(inputFilename=options.infile)
        doc.mdCommandLine = metadata.fromCommandLine(extras)
        doc.preprocess()
        doc.printTargets()
    elif options.jsonCode:
        doc = Spec(inputFilename=options.infile)
        doc.mdCommandLine = metadata.fromCommandLine(extras)
        doc.preprocess()
        exec(f"print(config.printjson({options.jsonCode}))")
    elif options.code:
        doc = Spec(inputFilename=options.infile)
        doc.mdCommandLine = metadata.fromCommandLine(extras)
        doc.preprocess()
        exec(f"print({options.code})")
    elif options.refreshData:
        m.state.printOn = "everything"
        update.updateReadonlyDataFiles()
        m.warn("Don't forget to bump the version number!")
    elif options.printMetadata:
        doc = Spec(inputFilename=options.infile)
        doc.mdCommandLine = metadata.fromCommandLine(extras)
        doc.preprocess()
        md = {
            "defaults.include": doc.mdDefaults.allData if doc.mdDefaults else [],
            "computed-metadata.include": doc.mdOverridingDefaults.allData if doc.mdOverridingDefaults else [],
            "document": doc.mdDocument.allData if doc.mdDocument else [],
            "command-line": doc.mdCommandLine.allData,
        }
        print(json.dumps(md, indent=2, default=printjson.getjson))  # noqa: T201


def handleRefs(options: argparse.Namespace, extras: list[str]) -> None:
    from . import datablocks, metadata
    from .refs import ReferenceManager
    from .Spec import Spec

    m.state.dieOn = "nothing"
    m.state.printOn = "nothing"
    doc = Spec(inputFilename=options.infile)
    if doc.valid:
        doc.mdCommandLine = metadata.fromCommandLine(extras)
        doc.preprocess()
        rm = doc.refs
    else:
        rm = ReferenceManager()
        rm.initializeRefs(datablocks=datablocks)
    if options.text:
        options.text = options.text
    refs = rm.queryAllRefs(
        text=options.text,
        linkFor=options.linkFor,
        linkType=options.linkType,
        status=options.status,
        spec=options.spec,
        latestOnly=options.latestOnly,
        exact=options.exact,
    )
    if m.state.printMode == "json":
        dumped = json.dumps(refs, indent=2, default=printjson.getjson)
        dumped = dumped[1:]  # Remove the opening [ since that's printed by printOpener()
        m.p(dumped)  # Go straight to m.p() so no formatting happens.
        sys.exit(0)  # Exit early so printCloser() doesn't run
    else:
        m.state.printOn = "everything"
        m.say(printjson.printjson(refs))


def handleIssuesList(options: argparse.Namespace) -> None:
    from . import issuelist

    if options.printTemplate:
        issuelist.printHelpMessage()
    else:
        issuelist.printIssueList(options.infile, options.outfile)


def handleSource(options: argparse.Namespace) -> None:
    from . import fonts

    try:
        fontPath = config.scriptPath("fonts", "smallblocks.bsfont")
        font = fonts.Font.fromPath(fontPath)
        fonts.replaceComments(font=font, inputFilename=options.infile, outputFilename=options.outfile)
    except Exception as e:
        m.die(f"Error trying to embiggen text:\n{e}")
        return


def handleOutline(options: argparse.Namespace) -> None:
    from . import outline
    from .Spec import Spec

    doc = Spec(inputFilename=options.infile)
    if not doc.valid:
        return
    with m.messagesSilent() as _:
        doc.preprocess()
    m.say(outline.printOutline(doc))


def handleTest(options: argparse.Namespace, extras: list[str]) -> None:
    from . import metadata, test

    md = metadata.fromCommandLine(extras)
    m.state.dieOn = "nothing"
    filters = test.TestFilter.fromOptions(options)
    if options.rebase:
        test.rebase(filters, md=md)
    else:
        result = test.run(filters, md=md)
        sys.exit(0 if result else 1)


def handleProfile(options: argparse.Namespace) -> None:
    root = f'--root="{options.root}"' if options.root else ""
    leaf = f'--leaf="{options.leaf}"' if options.leaf else ""
    if options.svgFile:
        os.system(  # noqa: S605
            f"time python -m cProfile -o stat.prof -m bikeshed -f spec && gprof2dot -f pstats --skew=.0001 {root} {leaf} stat.prof | dot -Tsvg -o {options.svgFile} && rm stat.prof",
        )
    else:
        os.system(  # noqa: S605
            f"time python -m cProfile -o /tmp/stat.prof -m bikeshed -f spec && gprof2dot -f pstats --skew=.0001 {root} {leaf} /tmp/stat.prof | xdot &",
        )


def handleTemplate(options: argparse.Namespace) -> None:
    if m.wrappedOutput():
        m.die(f"`bikeshed template` only supports printing to console. You set --print={m.state.printMode}")
        return
    if options.variant == "base":
        m.say(
            """<pre class='metadata'>
Title: Your Spec Title
Shortname: your-spec
Level: 1
Status: w3c/UD
Group: WGNAMEORWHATEVER
Repository: org/repo-name
URL: http://example.com/url-this-spec-will-live-at
Editor: Your Name, Your Company http://example.com/your-company, your-email@example.com, http://example.com/your-personal-website
Abstract: A short description of your spec, one or two sentences.
Complain About: accidental-2119 yes, missing-example-ids yes
Markup Shorthands: markdown yes, css no
</pre>

Introduction {#intro}
=====================

Introduction here.
""",
        )
    elif options.variant == "minimal":
        m.say(
            """<pre class='metadata'>
Title: Test
Editor: test
Abstract: test
Group: test
Status: ls
Shortname: test
Boilerplate: style-autolinks off, style-colors off, style-counters off, style-selflinks off, style-issues off, style-md-lists off, style-dfn-panel off, script-dfn-panel off, script-dfn-panel-json off, script-dom-helper off, style-ref-hints off, script-ref-hints off, script-link-titles off, table-of-contents off, abstract off, index off, references off
Markup Shorthands: markdown on
</pre>
""",
        )
    elif options.variant == "test":
        m.say(
            """<pre class=metadata>
Group: test
Shortname: foo
Level: 1
Status: LS
ED: http://example.com/foo
Editor: Example Editor
Date: 1970-01-01
Title: Test Title
Abstract: Test Description
</pre>
""",
        )


def handleWpt(options: argparse.Namespace) -> None:
    if m.wrappedOutput():
        m.die(f"`bikeshed wpt` only supports printing to console. You set --print={m.state.printMode}")
        return
    if options.template:
        m.say(
            """<!DOCTYPE html>
<meta charset=utf-8>
<title>window.offscreenBuffering</title>
<link rel=author title="AUTHOR NAME HERE" href="mailto:AUTHOR EMAIL HERE">
<link rel=help href="LINK TO ROUGHLY WHAT'S BEING TESTED HERE">
<script src="/resources/testharness.js"></script>
<script src="/resources/testharnessreport.js"></script>
<script>
/* Choose the test type you want: */


/* 1. Standard, synchronous test */
test(function() {
    /* test code here */
}, "TEST NAME HERE / SHORT DESCRIPTION PHRASE");


/* 2. Async test */
let t = async_test("TEST NAME HERE / SHORT DESCRIPTION PHRASE");
somethingWithACallback( function(){ t.step(()=>{ /* test code here */}) );
something.addEventListener('foo', t.step_func(()=>{ /* test code here */}));
t.done(); // when all tests are finished running
// or call the following if there's only one test, automatically does .done() for you
something.addEventListener('foo', t.step_func_done(()=>{ /* test code here */}));


/* 3. Promise test */
promise_test(function(){
    return somePromiseFunc().then(()=>{ /* test code here */ });
}, "TEST NAME HERE / SHORT DESCRIPTION PHRASE");
// auto-finishes when the returned promise fulfills
// or if promise should reject:
promise_test(function(t){
    return promise_rejects(t, new ExpectedError(), somePromiseCode());
}, "TEST NAME HERE / SHORT DESCRIPTION PHRASE");


/* "test code here" Asserts */
// Only use inside of /* test code here */ regions
assert_true(VALUE HERE, "TEST DESCRIPTION");
assert_equals(ACTUAL VALUE HERE, EXPECTED VALUE HERE, "TEST DESCRIPTION");
// lots more at http://web-platform-tests.org/writing-tests/testharness-api.html#list-of-assertions
</script>
""",
        )



from __future__ import annotations

import collections.abc
import hashlib
import html
import re
from collections import OrderedDict

import html5lib
from lxml import etree
from lxml.cssselect import CSSSelector
from lxml.html import tostring

from .. import t
from ..messages import die, warn

if t.TYPE_CHECKING:
    ElementPredT: t.TypeAlias = t.Callable[[t.ElementT], bool]


def flatten(arr: t.Iterable) -> t.Iterator:
    for el in arr:
        if isinstance(el, collections.abc.Iterable) and not isinstance(el, str) and not etree.iselement(el):
            yield from flatten(el)
        else:
            yield el


def unescape(string: str) -> str:
    return html.unescape(string)


def findAll(sel: str, context: t.SpecT | t.ElementT | t.DocumentT) -> list[t.ElementT]:
    context = t.cast("t.ElementT", getattr(context, "document", context))
    try:
        return t.cast("list[t.ElementT]", CSSSelector(sel, namespaces={"svg": "http://www.w3.org/2000/svg"})(context))
    except Exception as e:
        die(f"The selector '{sel}' returned an error:\n{e}")
        return []


def find(sel: str, context: t.SpecT | t.ElementT | t.DocumentT) -> t.ElementT | None:
    result = findAll(sel, context)
    if result:
        return result[0]
    else:
        return None


def escapeCSSIdent(val: str) -> str:
    if len(val) == 0:
        die("Programming error: can't escape an empty ident.")
        return ""
    ident = ""
    firstCode = val[0]
    for i, code in enumerate(ord(x) for x in val):
        if code == 0:
            die(f"Invalid character: the string '{val}' somehow has a NUL in it.")
            return ""
        if (
            0x1 <= code <= 0x1F
            or code == 0x7F
            or (i == 0 and 0x30 <= code <= 0x39)
            or (i == 1 and 0x30 <= code <= 0x39 and firstCode == 0x2D)
        ):
            ident += rf"\{code:x} "
        elif (
            code >= 0x80 or code in (0x2D, 0x5F) or 0x30 <= code <= 0x39 or 0x41 <= code <= 0x5A or 0x61 <= code <= 0x7A
        ):
            ident += chr(code)
        else:
            ident += r"\{}".format(chr(code))
    return ident


def escapeUrlFrag(val: str) -> str:
    result = ""
    for char in val:
        if validUrlUnit(char):
            result += char
        else:
            for b in char.encode("utf-8"):
                result += f"%{b:0>2x}"
    return result


def validUrlUnit(char: str) -> bool:
    c = ord(char)
    if c < 0xA0:
        # ASCII range
        return (
            c in (0x21, 0x24)
            or 0x26 <= c <= 0x29
            or 0x2A <= c <= 0x3B
            or c == 0x3D
            or 0x3F <= c <= 0x5A
            or c == 0x5F
            or 0x61 <= c <= 0x7A
            or c == 0x7E
        )
    else:
        if 0xD800 <= c <= 0xDFFF or 0xFDD0 <= c <= 0xFDEF:
            return False
        if (c % 0xFFFF) in [0xFFFE, 0xFFFF]:  # noqa: SIM103
            # Last two bytes are FFFE or FFFF
            return False
        return True


def textContent(el: t.ElementT, exact: bool = False) -> str:
    # If exact is False, then any elements with data-deco attribute
    # get ignored in the textContent.
    # This allows me to ignore things added by Bikeshed by default.
    if len(el) == 0:
        return el.text or ""
    if exact:
        return t.cast(str, tostring(el, method="text", with_tail=False, encoding="unicode"))
    else:
        return textContentIgnoringDecorative(el)


def textContentIgnoringDecorative(el: t.ElementT) -> str:
    s = ""
    for child in childNodes(el):
        if isinstance(child, str):
            s += child
        elif child.get("data-deco") is None:
            s += textContentIgnoringDecorative(child)
    return s


def innerHTML(el: t.ElementT | None) -> str:
    if el is None:
        return ""
    return (el.text or "") + "".join(tostring(x, encoding="unicode") for x in el)


def outerHTML(el: t.NodesT | None, literal: bool = False, with_tail: bool = False) -> str:
    if el is None:
        return ""
    if isinstance(el, str):
        return el
    if isinstance(el, list):
        return "".join(outerHTML(x) for x in el)
    if el.get("bs-autolink-syntax") is not None and not literal:
        return el.get("bs-autolink-syntax") or ""
    return t.cast(str, tostring(el, with_tail=with_tail, encoding="unicode"))


def printNodeTree(node: t.NodeT | str) -> str:
    # Debugging tool
    if isinstance(node, str):
        return "#text: " + repr(node)
    if isOddNode(node):
        return outerHTML(node)
    if isinstance(node, list):
        s = "[]"
    else:
        s = f"{serializeTag(node)}"
    linesPerChild = [printNodeTree(child).split("\n") for child in childNodes(node, skipOddNodes=False)]
    if linesPerChild:
        for childLines in linesPerChild[:-1]:
            childLines[0] = " ├" + childLines[0]
            childLines[1:] = [" │" + line for line in childLines[1:]]
            s += "\n" + "\n".join(childLines)
        childLines = linesPerChild[-1]
        childLines[0] = " ╰" + childLines[0]
        childLines[1:] = ["  " + line for line in childLines[1:]]
        s += "\n" + "\n".join(childLines)
    return s


def linkTextsFromElement(el: t.ElementT) -> list[str]:
    texts: list[str] = []
    if el.get("data-lt") == "":
        return []
    elif el.get("data-lt"):
        rawText = el.get("data-lt", "")
        if rawText in ["|", "||", "|||"]:
            texts = [rawText]
        else:
            texts = [x.strip() for x in rawText.split("|")]
    else:
        if el.tag in ("dfn", "a"):
            texts = [textContent(el).strip()]
        elif el.tag in ("h2", "h3", "h4", "h5", "h6"):
            textEl = find(".content", el)
            if textEl is None:
                textEl = el
            texts = [textContent(textEl).strip()]
    if el.get("data-local-lt"):
        localTexts = [x.strip() for x in el.get("data-local-lt", "").split("|")]
        for text in localTexts:
            if text in texts:
                # lt and local-lt both specify the same thing
                raise DuplicatedLinkText(text, texts + localTexts, el)
        texts += localTexts

    texts = [re.sub(r"\s+", " ", x) for x in texts if x != ""]
    return texts


class DuplicatedLinkText(Exception):
    def __init__(self, offendingText: str, allTexts: list[str], el: t.ElementT) -> None:
        super().__init__()
        self.offendingText = offendingText
        self.allTexts = allTexts
        self.el = el

    def __unicode__(self) -> str:
        return f"<Text '{self.offendingText}' shows up in both lt and local-lt>"


def firstLinkTextFromElement(el: t.ElementT) -> str | None:
    try:
        texts = linkTextsFromElement(el)
    except DuplicatedLinkText as e:
        texts = e.allTexts
    return texts[0] if len(texts) > 0 else None


def serializeTag(el: t.ElementT) -> str:
    # Serialize *just* the opening tag for the element.
    # Use when you want to output the HTML,
    # but it might be a container with a lot of content.
    tag = "<" + el.tag
    for n, v in sorted(el.attrib.items()):
        # Don't output the bs-* attributes, they're added by BS
        # and don't show up in the source, so it's confusing to
        # print them.
        if t.cast(str, n).startswith("bs-"):
            continue
        tag += ' {n}="{v}"'.format(n=str(n), v=escapeAttr(str(v)))
    tag += ">"
    return tag


def tagName(el: t.ElementT | None) -> str | None:
    # Returns the tagname, or None if passed None
    # Iow, safer version of el.tagName
    if el is None:
        return None
    return el.tag


def foldWhitespace(text: str) -> str:
    return re.sub(r"(\s|\xa0)+", " ", text)


def sortElements(el: t.Iterable[t.ElementT]) -> list[t.ElementT]:
    return sorted(el, key=lambda x: (x.get("bs-line-number", ""), textContent(x)))


def parseHTML(text: str) -> list[t.ElementT]:
    doc = html5lib.parse(text, treebuilder="lxml", namespaceHTMLElements=False)
    head = doc.getroot()[0]
    body = doc.getroot()[1]
    if len(body) > 0 or body.text is not None:
        # Body contains something, so return that.
        contents = [body.text] if body.text is not None else []
        contents.extend(childElements(body))
        return contents
    elif len(head) > 0 or head.text is not None:
        # Okay, anything in the head?
        contents = [head.text] if head.text is not None else []
        contents.extend(childElements(head))
        return contents
    else:
        return []


def parseDocument(text: str) -> t.DocumentT:
    doc = html5lib.parse(text, treebuilder="lxml", namespaceHTMLElements=False)
    return t.cast("t.DocumentT", doc)


def escapeHTML(text: str) -> str:
    # Escape HTML
    return text.replace("&", "&amp;").replace("<", "&lt;")


def escapeAttr(text: str) -> str:
    return text.replace("&", "&amp;").replace("'", "&apos;").replace('"', "&quot;")


def clearContents(el: t.ElementT) -> t.ElementT:
    del el[:]
    el.text = ""
    return el


def parentElement(el: t.ElementT | None, depth: int = 1) -> t.ElementT | None:
    for _ in range(depth):
        if el is None:
            return None
        el = el.getparent()
    return el


def nextSiblingNode(el: t.ElementT) -> t.ElementT | None:
    return el.getnext()


def nextSiblingElement(el: t.ElementT) -> t.ElementT | None:
    while True:
        next = nextSiblingNode(el)
        if next is None:
            return None
        if isElement(next):
            return next


@t.overload
def appendChild(parent: t.ElementT, *els: t.NodesT, allowEmpty: t.Literal[False] = False) -> t.ElementT: ...


@t.overload
def appendChild(parent: t.ElementT, *els: t.NodesT, allowEmpty: bool) -> t.ElementT | None: ...


def appendChild(parent: t.ElementT, *els: t.NodesT, allowEmpty: bool = False) -> t.ElementT | None:
    # Appends either text or an element.
    child: t.NodeT | None = None
    parentLen = len(parent)
    for child in flatten(els):
        assert child is not None
        if isinstance(child, str):
            if parentLen > 0:
                parent[-1].tail = (parent[-1].tail or "") + child
            else:
                parent.text = (parent.text or "") + child
        else:
            parentLen += 1
            if parentLen == 0 and parent.text is not None:
                # LXML "helpfully" assumes you meant to insert it before the text,
                # and so moves the text into the element's tail when you append.
                text, parent.text = parent.text, None
                parent.append(child)
                parent.text = text
            else:
                # For some reason it doesn't make any weird assumptions about text
                # when the parent already has children; the last child's tail
                # doesn't get moved into the appended child or anything.
                parent.append(child)
    if child is None and not allowEmpty:
        msg = "Empty child list appended without allowEmpty=True"
        raise Exception(msg)
    if isElement(child):
        return child
    else:
        return None


def prependChild(parent: t.ElementT, *children: t.NodesT) -> None:
    # Prepends either text or an element to the parent.
    for child in reversed(list(flatten(children))):
        if isinstance(child, str):
            if parent.text is None:
                parent.text = child
            else:
                parent.text = child + parent.text
        else:
            removeNode(child)
            parent.insert(0, child)
            if parent.text is not None:
                child.tail = (child.tail or "") + parent.text
                parent.text = None


def insertBefore(target: t.ElementT, *els: t.NodesT) -> t.ElementT:
    parent = target.getparent()
    assert parent is not None
    index = parent.index(target)
    prevSibling = parent[index - 1] if index > 0 else None
    for el in flatten(els):
        if isinstance(el, str):
            if prevSibling is not None:
                prevSibling.tail = (prevSibling.tail or "") + el
            else:
                parent.text = (parent.text or "") + el
        else:
            parent.insert(index, el)
            index += 1
            prevSibling = el
    return target


def insertAfter(target: t.ElementT, *els: t.NodesT) -> t.ElementT:
    parent = target.getparent()
    assert parent is not None
    for el in flatten(els):
        if isinstance(el, str):
            target.tail = (target.tail or "") + el
        else:
            parent.insert(parent.index(target) + 1, el)
            target = el
    return target


def removeNode(node: t.ElementT) -> t.ElementT:
    # Kills the node *and* its children.
    # If you just want to remove the node itself, and lift
    # the contents up into its place, you wanna call
    # replaceWithContents()
    parent = node.getparent()
    if parent is None:
        return node
    index = parent.index(node)
    text = node.tail or ""
    node.tail = None
    if index == 0:
        parent.text = (parent.text or "") + text
    else:
        prevsibling = parent[index - 1]
        prevsibling.tail = (prevsibling.tail or "") + text
    parent.remove(node)
    return node


def replaceNode(node: t.ElementT, *replacements: t.NodesT) -> t.NodesT | None:
    insertBefore(node, *replacements)
    removeNode(node)
    if replacements:
        return replacements[0]
    return None


def transferAttributes(source: t.ElementT, target: t.ElementT) -> t.ElementT:
    for k, v in source.attrib.items():
        target.set(k, v)
    return target


def appendContents(el: t.ElementT, container: t.ElementT | t.Iterable[t.NodesT]) -> t.ElementT:
    # Accepts either an iterable *or* a container element
    if isElement(container):
        container = childNodes(container, clear=True)
    appendChild(el, *container, allowEmpty=True)
    return el


def replaceContents(el: t.ElementT, newElements: t.NodesT | t.Iterable[t.NodesT]) -> t.ElementT:
    clearContents(el)
    return appendContents(el, newElements)


def replaceWithContents(el: t.ElementT) -> t.NodesT | None:
    return replaceNode(el, childNodes(el, clear=True))


def moveContents(toEl: t.ElementT, fromEl: t.ElementT) -> None:
    replaceContents(toEl, fromEl)
    fromEl.text = ""


def wrapContents(parentEl: t.ElementT, wrapperEl: t.ElementT) -> t.ElementT:
    appendContents(wrapperEl, parentEl)
    appendChild(parentEl, wrapperEl, allowEmpty=True)
    return parentEl


def headingLevelOfElement(el: t.ElementT) -> str | None:
    for heading in relevantHeadings(el, levels=[2, 3, 4, 5, 6]):
        if heading.get("data-level") is not None:
            return heading.get("data-level")
    return None


def relevantHeadings(startEl: t.ElementT, levels: list[int] | None = None) -> t.Generator[t.ElementT, None, None]:
    if levels is None:
        levels = [1, 2, 3, 4, 5, 6]
    tagNames = ["h" + str(level) for level in levels]
    currentHeadingLevel = float("inf")
    for el in scopingElements(startEl, tagNames):
        tagLevel = int(el.tag[1])
        if tagLevel < currentHeadingLevel:
            yield el
            currentHeadingLevel = tagLevel
        if tagLevel == 2:
            return


def sectionName(doc: t.SpecT, el: t.ElementT) -> str | None:
    """
    Return the name of the nearest section to el,
    or None if that section isn't meant to be referenced.
    """
    try:
        h = next(relevantHeadings(el))
    except StopIteration:
        return "Unnamed section"
    if hasClass(doc, h, "no-ref"):
        return None
    return textContent(h)


def collectLinksWithSectionNames(
    doc: t.SpecT,
    root: t.ElementT | None = None,
    links: dict[t.ElementT, str | None] | None = None,
    name: str | None = "Unnamed section",
) -> dict[t.ElementT, str | None]:
    # Tree-walk to collect all links, and compute their section names
    # along the way.
    if links is None:
        links = {}
    if root is None:
        root = doc.body
    for child in childElements(root):
        if tagName(child) == "a":
            links[child] = name
            continue
        if tagName(child) in ("h1", "h2", "h3", "h4", "h5", "h6"):
            if hasClass(doc, child, "no-ref"):
                name = None
            else:
                name = textContent(child)
        collectLinksWithSectionNames(doc, child, links, name)
    return links


def scopingElements(startEl: t.ElementT, tags: list[str]) -> t.Generator[t.ElementT, None, None]:
    # Elements that could form a "scope" for the startEl
    # Ancestors, and preceding siblings of ancestors.
    # Maps to the things that can establish a counter scope.
    tagFilter = set(tags)

    for sib in siblingElements(startEl, preceding=True):
        if sib.tag in tagFilter:
            yield sib
    for ancestor in ancestorElements(startEl):
        if ancestor.tag in tagFilter:
            yield ancestor
        for sib in siblingElements(ancestor, preceding=True):
            if sib.tag in tagFilter:
                yield sib


def previousElements(startEl: t.ElementT, tag: str | None = None, *tags: str) -> list[t.ElementT]:
    # Elements preceding the startEl in document order.
    # Like .iter(), but in the opposite direction.
    els: list[t.ElementT] = []
    for el in startEl.getroottree().getroot().iter(tag, *tags):
        if el == startEl:
            return list(reversed(els))
        els.append(el)
    return els


def childElements(parentEl: t.ElementT, oddNodes: bool = False) -> t.Generator[t.ElementT, None, None]:
    tag = None if oddNodes else "*"
    yield from parentEl.iterchildren(tag)


def siblingElements(el: t.ElementT, preceding: bool = False) -> t.Iterable[t.ElementT]:
    return el.itersiblings("*", preceding=preceding)


def ancestorElements(el: t.ElementT, self: bool = False) -> t.Generator[t.ElementT, None, None]:
    if self:
        yield el
    yield from el.iterancestors()


def childNodes(
    parentEl: t.ElementishT,
    clear: bool = False,
    skipOddNodes: bool = True,
    mergeText: bool = False,
) -> list[t.NodeT]:
    """
    This function returns all the nodes in a parent element in the DOM sense,
    mixing text nodes (strings) and other nodes together
    (rather than LXML's default stupid handling of text).

    If you set "clear" to True, it'll
    1. remove all of parentEl's children,
       so you can append nodes back to it safely, and
    2. Set parentEl.text and child elements' .tail to null,
       again so you can safely append text to parentEl.
    In other words, the following is a no-op:

    ```
    appendChild(parentEl, *childNodes(parentEl, clear=True, skipOddNodes=False), allowEmpty=True)
    ```

    Using clear=True is required if you're going to be modifying the element or its children,
    otherwise you'll get weird results (repeated/misplaced text).
    But if you're just reading nodes,
    it's not necessary.

    skipOddNodes ensures that the return value will only be text and Element nodes;
    if it's false, there might be comments, PIs, etc.

    mergeText merges adjacent text nodes in the output, even if they weren't adjacent
    in the source document (such as if a skipped odd node separated them).
    """
    ret: list[t.NodeT] = []

    def append(nodes: list[t.NodeT], node: t.NodeT) -> None:
        if not mergeText:
            nodes.append(node)
            return
        if isinstance(node, str) and len(nodes) > 0 and isinstance(nodes[-1], str):
            nodes[-1] += node
        else:
            nodes.append(node)

    if isinstance(parentEl, list):
        for c in parentEl:
            if isinstance(c, str):
                append(ret, c)
                continue
            if skipOddNodes and isOddNode(c):
                pass
            else:
                append(ret, c)
            if not emptyText(c.tail, wsAllowed=False):
                append(ret, t.cast(str, c.tail))
                if clear:
                    c.tail = None
        if clear:
            parentEl[:] = []
        return ret

    if not emptyText(parentEl.text, wsAllowed=False):
        append(ret, t.cast(str, parentEl.text))
        if clear:
            parentEl.text = None
    for c in childElements(parentEl, oddNodes=True):
        if skipOddNodes and isOddNode(c):
            pass
        else:
            append(ret, c)
        if not emptyText(c.tail, wsAllowed=False):
            append(ret, t.cast(str, c.tail))
            if clear:
                c.tail = None
    if clear:
        clearContents(parentEl)
    return ret


def nodeIter(el: t.ElementT, clear: bool = False, skipOddNodes: bool = True) -> t.Generator[t.NodeT, None, None]:
    # Iterates thru an element and all its descendants,
    # yielding up each child node it sees in depth-first order.
    # (In other words, same as el.iter(),
    #  but returning nodes+strings rather than the stupid LXML model.)
    # Takes the same kwargs as childNodes
    if isinstance(el, str):
        yield el
        return
    if isinstance(el, etree.t.ElementTree):
        el = el.getroot()
    text = el.text
    tail = el.tail
    if clear:
        el.text = None
        el.tail = None
    yield el
    if text is not None:
        yield text
    for c in childElements(el, oddNodes=True):
        if skipOddNodes and isOddNode(c):
            continue
        # yield from nodeIter(c, clear=clear, skipOddNodes=skipOddNodes)
        yield from nodeIter(c, clear=clear, skipOddNodes=skipOddNodes)
    if tail is not None:
        yield tail


def treeAttr(el: t.ElementT, attrName: str) -> str | None:
    # Find the nearest instance of the given attr in the tree
    # Useful for when you can put an attr on an ancestor and apply it to all contents.
    # Returns attrValue or None if nothing is found.

    for target in ancestorElements(el, self=True):
        if target.get(attrName) is not None:
            return target.get(attrName)
    return None


def closestAttr(el: t.ElementT, *attrs: str) -> tuple[str, str] | tuple[None, None]:
    # Like treeAttr, but can provide multiple attr names, and returns the first one found.
    # Useful with combos like highlight/nohighlight
    # If multiple target attrs show up on same element, priority is calling order.
    # Returns a tuple of (attrName, attrValue) or (None, None) if nothing is found.

    for target in ancestorElements(el, self=True):
        for attrName in attrs:
            if target.get(attrName) is not None:
                return attrName, t.cast(str, target.get(attrName))
    return None, None


def closestAncestor(el: t.ElementT, pred: ElementPredT) -> t.ElementT | None:
    # Finds the nearest ancestor matching a predicate
    for target in ancestorElements(el):
        if pred(target):
            return target
    return None


def filterAncestors(el: t.ElementT, pred: ElementPredT) -> t.Generator[t.ElementT, None, None]:
    # Returns all ancestors that match the predicate
    for target in el.iterancestors():
        if pred(target):
            yield target


def hasAncestor(el: t.ElementT, pred: ElementPredT) -> bool:
    return closestAncestor(el, pred) is not None


def removeAttr(el: t.ElementT, *attrNames: str) -> t.ElementT:
    # Remove an attribute, silently ignoring if attr doesn't exist.
    for attrName in attrNames:
        if attrName in el.attrib:
            del el.attrib[attrName]
    return el


def renameAttr(el: t.ElementT, oldAttr: str, newAttr: str) -> t.ElementT:
    # If oldAttr exists on the element, moves its value to newAttr
    # and deletes oldAttr. Silently does nothing if oldAttr doesn't exist.
    val = el.get(oldAttr)
    if val is not None:
        el.set(newAttr, val)
        removeAttr(el, oldAttr)
    return el


def hasAnyAttr(el: t.ElementT, *attrNames: str) -> bool:
    # Returns True if the element has at least one of the named attributes
    return any(attrName in el.attrib for attrName in attrNames)


def hasAttr(el: t.ElementT, attrName: str) -> bool:
    return attrName in el.attrib


def hasAttrs(el: t.ElementT) -> bool:
    return bool(el.attrib)


def addClass(doc: t.SpecT, el: t.ElementT, cls: str) -> t.ElementT:
    if el.get("class") is None:
        el.set("class", cls)
    elif hasClass(doc, el, cls):
        pass
    else:
        el.set("class", "{} {}".format(el.get("class"), cls))
    return el


def hasClass(doc: t.SpecT, el: t.ElementT, cls: str) -> bool:
    elClass = el.get("class")
    if elClass is None:
        return False
    if cls == elClass:
        return True
    if cls not in elClass:
        return False
    key = cls, elClass
    if key in doc.cachedClassTests:
        return doc.cachedClassTests[key]
    ret = bool(re.search(r"(^|\s)" + cls + r"($|\s)", elClass))
    doc.cachedClassTests[key] = ret
    return ret


def removeClass(el: t.ElementT, cls: str) -> t.ElementT:
    oldClass = el.get("class")
    if oldClass is None:
        return el
    newClass = " ".join(c for c in oldClass.split() if c != cls)
    if newClass == "":
        del el.attrib["class"]
    else:
        el.set("class", newClass)
    return el


def isElement(node: t.Any) -> t.TypeGuard[t.ElementT]:
    # LXML HAS THE DUMBEST XML TREE DATA MODEL IN THE WORLD
    return etree.iselement(node) and isinstance(node.tag, str)


def isNode(node: t.Any) -> t.TypeGuard[t.NodeT]:
    return isElement(node) or isinstance(node, str)


def isNodes(nodes: t.Any) -> t.TypeGuard[t.NodesT]:
    if isNode(nodes):
        return True
    if not isinstance(nodes, list):
        return False
    return all(isNodes(child) for child in nodes)


def isOddNode(node: t.Any) -> bool:
    # Something other than an element node or string.
    if isinstance(node, str):
        return False
    if isElement(node):  # noqa: SIM103
        return False
    return True


def isNormative(doc: t.SpecT, el: t.ElementT) -> bool:
    # Returns whether the element is "informative" or "normative" with a crude algo.
    # Currently just tests whether the element is in a class=example or class=note block, or not.
    if el in doc.cachedNormativeEls:
        return doc.cachedNormativeEls[el]
    informativeClasses = [
        "note",
        "example",
        "non-normative",
        "informative",
    ] + doc.md.informativeClasses
    for cls in informativeClasses:
        if hasClass(doc, el, cls):
            doc.cachedNormativeEls[el] = False
            return False
    if hasClass(doc, el, "normative"):
        doc.cachedNormativeEls[el] = True
        return True
    parent = parentElement(el)
    if not isElement(parent):
        # Went past the root without finding any indicator,
        # so normative by default.
        doc.cachedNormativeEls[el] = True
        return True
    # Otherwise, walk the tree
    norm = isNormative(doc, parent)
    doc.cachedNormativeEls[el] = norm
    return norm


def isEmpty(el: t.ElementT) -> bool:
    # Returns whether the element is empty - no text or children.
    return (el.text is None or el.text.strip() == "") and len(el) == 0


def hasChildElements(el: t.ElementT) -> bool:
    try:
        next(childElements(el))
        return True
    except StopIteration:
        return False


# If the element has one child element, returns it.
# Otherwise, returns None
def hasOnlyChild(el: t.ElementT, wsAllowed: bool = True) -> t.ElementT | None:
    if not emptyText(el.text, wsAllowed):
        # Has significant child text
        return None
    children = childElements(el)
    single = next(children, None)
    if single is None:
        # No children
        return None
    if not emptyText(single.tail, wsAllowed):
        # Has significant child text following the child element
        return None
    if next(children, None) is not None:
        # At least two children
        return None
    return single


def isOnlyChild(el: t.ElementT, wsAllowed: bool = True) -> bool:
    parent = parentElement(el)
    if parent is None:
        return True
    return hasOnlyChild(parent, wsAllowed) is not None


def fixSurroundingTypography(el: t.ElementT) -> t.ElementT:
    # Applies some of the fixTypography changes to the content surrounding an element.
    # Used when a shorthand prevented fixTypography from firing previously.
    if el.tail is not None and el.tail.startswith("'"):
        el.tail = "’" + el.tail[1:]
    return el


def unfixTypography(text: str) -> str:
    # Replace curly quotes with straight quotes, and emdashes with double dashes.
    text = re.sub(r"’", r"'", text)
    # Fix line-ending em dashes, or --, by moving the previous line up, so no space.
    text = re.sub(r"—<wbr>", r"--", text)
    return text


def emptyText(text: str | None, wsAllowed: bool = True) -> bool:
    # Because LXML represents a complete lack of text as None,
    # you can't do something like `el.text.strip() == ""` to test for emptiness.
    # wsAllowed controls whether whitespace-only strings count as empty or not
    if text is None:
        return True
    if not wsAllowed:
        return text == ""
    return text.strip() == ""


def hashContents(el: t.ElementT) -> str:
    # Hash the contents of an element into an 8-character alphanum string.
    # Generally used for generating probably-unique IDs.
    # Normalize whitespace away to avoid git-related newline normalization issues.
    text = re.sub(r"\s+", " ", textContent(el).strip()).encode("ascii", "xmlcharrefreplace")
    return hashlib.md5(text).hexdigest()[0:8]


def replaceMacrosTextly(text: str, macros: t.Mapping[str, str], context: str) -> str:
    # Same as replaceMacros(), but does the substitution
    # directly on the text, rather than relying on the
    # html parser to have preparsed the macro syntax
    def macroReplacer(match: re.Match) -> str:
        fullText = t.cast(str, match.group(0))
        innerText = match.group(2).lower() or ""
        optional = match.group(3) == "?"
        if fullText.startswith("\\"):
            # Escaped
            return fullText[1:]
        if fullText.startswith("[["):
            # Actually a biblio link
            return fullText
        if innerText in macros:
            # For some reason I store all the macros in lowercase,
            # despite requiring them to be spelled with uppercase.
            return macros[innerText]
        # Nothing has matched, so start failing the macros.
        if optional:
            return ""
        die(
            f"Found unmatched text macro {fullText}. Correct the macro, or escape it with a leading backslash.",
            lineNum=context,
        )
        return fullText

    return re.sub(r"(\\|\[)?\[([A-Z\d-]*[A-Z][A-Z\d-]*)(\??)\]", macroReplacer, text)


def fixupIDs(doc: t.SpecT, els: t.Iterable[t.ElementT]) -> None:
    addOldIDs(els)
    dedupIDs(doc)


def safeID(doc: t.SpecT, id: str) -> str:
    # Converts generated IDs into author-specified versions,
    # for when an ID isn't otherwise settable by the author.
    trans = doc.md.translateIDs
    if id in trans:
        return trans[id]
    return id


def uniqueID(*s: str) -> str:
    # Turns a unique string into a more compact (and ID-safe)
    # hashed string
    return hashlib.md5("".join(s).encode("utf-8")).hexdigest()[:8]


def addOldIDs(els: t.Iterable[t.ElementT]) -> None:
    for el in els:
        oldIdAttr = el.get("oldids")
        if not oldIdAttr:
            continue
        oldIDs = [id.strip() for id in oldIdAttr.strip().split(",")]
        for oldID in oldIDs:
            prependChild(el, E.span({"id": oldID, "class": "bs-old-id"}))
        removeAttr(el, "oldids")


def dedupIDs(doc: t.SpecT) -> None:
    ids: OrderedDict[str, list[t.ElementT]] = OrderedDict()
    for el in findAll("[id]", doc):
        ids.setdefault(t.cast(str, el.get("id")), []).append(el)
    for dupeId, els in list(ids.items()):
        if len(els) < 2:
            # Only one instance, so nothing to do.
            continue
        warnAboutDupes = True
        if re.match(r"issue-[0-9a-fA-F]{8}$", dupeId):
            # Don't warn about issues, it's okay if they have the same ID because they're identical text.
            warnAboutDupes = False
        if dupeId.startswith("ref-for-"):
            warnAboutDupes = False
        complaintEls = []
        if warnAboutDupes:
            for el in els:
                if el.get("data-silently-dedup") is not None:
                    continue
                complaintEls.append(el)
        # Now dedup everything left in the list after the first one.
        dedupIndex = 1
        for el in els[1:]:
            if el.get("data-alternate-id"):
                altId = el.get("data-alternate-id")
                assert altId is not None
                if altId not in ids:
                    el.set("id", safeID(doc, altId))
                    ids[altId] = [el]
                    complaintEls.remove(el)
                    continue
            altId = f"{dupeId}{circledDigits(dedupIndex)}"
            while altId in ids:
                dedupIndex += 1
                altId = f"{dupeId}{circledDigits(dedupIndex)}"
            el.set("id", safeID(doc, altId))
            ids[altId] = [el]

        if len(complaintEls) > 1:
            complaintDetails = [f"<{el.tag}> on line {approximateLineNumber(el)}" for el in complaintEls]
            warn(
                f"Multiple elements have the same id '{dupeId}':\n  {', '.join(complaintDetails)}\nDeduping, but this ID may not be stable across revisions.",
            )


def approximateLineNumber(el: t.ElementT, setIntermediate: bool = True) -> str | None:
    if el.get("bs-line-number"):
        return el.get("bs-line-number")
    parent = parentElement(el)
    if not isElement(parent):
        return None
    approx = approximateLineNumber(parent, setIntermediate=setIntermediate)
    if approx is None:
        return None
    if approx[0].isdigit():
        approx = "~" + approx
    if setIntermediate:
        el.set("bs-line-number", approx)
    return approx


def circledDigits(num: int) -> str:
    """
    Converts a base-10 number into a string using unicode circled digits.
    That is, 123 becomes "①②③"
    """
    assert num >= 0
    digits = ["⓪", "①", "②", "③", "④", "⑤", "⑥", "⑦", "⑧", "⑨"]
    result = "".join(digits[int(d)] for d in str(num))
    return result


def collectIds(el: t.ElementT, ids: dict[str, t.ElementT] | None = None) -> dict[str, t.ElementT]:
    if ids is None:
        ids = {}
    id = el.get("id", "")
    if id and id not in ids:
        ids[id] = el
    for child in childElements(el):
        collectIds(child, ids)
    return ids


def collectAutolinks(el: t.ElementT, links: list[t.ElementT] | None = None) -> list[t.ElementT]:
    if links is None:
        links = []
    if tagName(el) == "a" and not hasAttr(el, "href") and el.get("data-link-type", "") != "biblio":
        links.append(el)
    elif tagName(el) == "bs-link":
        links.append(el)
    for child in childElements(el):
        collectAutolinks(child, links)
    return links


def collectSyntaxHighlightables(el: t.ElementT, els: list[t.ElementT] | None = None) -> list[t.ElementT]:
    if els is None:
        els = []
    if tagName(el) in ("xmp", "pre", "code"):
        els.append(el)
    for child in childElements(el):
        collectSyntaxHighlightables(child, els)
    return els


def createElement(tag: str, attrs: t.Mapping[str, str | None] | None = None, *children: t.NodesT | None) -> t.ElementT:
    if attrs is None:
        attrs = {}
    el: t.ElementT = etree.Element(tag, {n: v for n, v in attrs.items() if v is not None})
    if children:
        appendChild(el, *(x for x in children if x is not None), allowEmpty=True)
    return el


if t.TYPE_CHECKING:

    class ElementCreatorFnT(t.Protocol):
        def __call__(
            self,
            attrsOrChild: t.Mapping[str, str | None] | t.NodesT | None = None,
            *children: t.NodesT | None,
        ) -> t.ElementT: ...


class ElementCreationHelper:
    def __getattr__(self, name: str) -> ElementCreatorFnT:
        def _creater(
            attrsOrChild: t.Mapping[str, str | None] | t.NodesT | None = None,
            *children: t.NodesT | None,
        ) -> t.ElementT:
            if isNodes(attrsOrChild):
                return createElement(name, None, attrsOrChild, *children)
            else:
                assert isinstance(attrsOrChild, dict) or attrsOrChild is None
                return createElement(name, attrsOrChild, *children)

        return _creater


E = ElementCreationHelper()



from __future__ import annotations

from collections import OrderedDict

from .. import h, t
from .. import messages as m
from ..translate import _t


def addDfnPanels(doc: t.SpecT, dfns: list[t.ElementT]) -> None:
    # Constructs "dfn panels" which show all the local references to a term
    # Gather all the <a href>s together
    linksFromTargetId: OrderedDict[str, list[t.ElementT]] = OrderedDict()
    sectionNameFromLink = h.collectLinksWithSectionNames(doc)
    for a in sectionNameFromLink:
        href = a.get("href")
        if href is None:
            continue
        if not href.startswith("#"):
            continue
        linksFromTargetId.setdefault(href[1:], []).append(a)
    panelsJSON = doc.extraJC.addDfnPanels()
    for dfn in dfns:
        id = dfn.get("id")
        dfnText = h.textContent(dfn)
        if not id:
            # Something went wrong, bail.
            continue
        refsFromSection: OrderedDict[str, list[t.ElementT]] = OrderedDict()
        for link in linksFromTargetId.get(id, []):
            sectionName = sectionNameFromLink.get(link)
            if sectionName is not None:
                refsFromSection.setdefault(sectionName, []).append(link)
        h.addClass(doc, dfn, "dfn-paneled")
        sectionsJson = []
        for text, els in refsFromSection.items():
            refsJson = []
            for el in els:
                refID = el.get("id")
                if refID is None:
                    refID = f"ref-for-{id}"
                    el.set("id", h.safeID(doc, refID))
                refsJson.append({"id": refID})
            sectionsJson.append(
                {
                    "refs": refsJson,
                    "title": text,
                },
            )
        panelsJSON[id] = {
            "dfnID": id,
            "url": "#" + h.escapeUrlFrag(id),
            "dfnText": dfnText,
            "refSections": sectionsJson,
            "external": False,
        }


def addExternalDfnPanel(termEl: t.ElementT, ref: t.RefWrapper, doc: t.SpecT) -> None:
    # Constructs "dfn panels" which show all the local references to an external term

    # Calculate and cache the doc's links,
    # so I'm not iterating the doc for links constantly.
    if not doc.cachedLinksFromHref:
        for a in h.findAll("a", doc):
            href = a.get("href")
            if href is None:
                continue
            if href.startswith("#"):
                continue
            doc.cachedLinksFromHref.setdefault(href, []).append(a)

    if ref.url not in doc.cachedLinksFromHref:
        return

    panelsJSON = doc.extraJC.addDfnPanels()

    # Group the relevant links according to the section they're in.
    refsFromSection: OrderedDict[str, list[t.ElementT]] = OrderedDict()
    for link in doc.cachedLinksFromHref[ref.url]:
        section = h.sectionName(doc, link) or _t("Unnumbered Section")
        refsFromSection.setdefault(section, []).append(link)

    h.addClass(doc, termEl, "dfn-paneled")
    termID = termEl.get("id")
    if termID is None:
        m.warn("An external reference index entry ended up without an ID:\n{ref}")
        return
    termText = h.textContent(termEl)
    sectionsJson = []
    counter = 0
    for text, els in refsFromSection.items():
        refsJson = []
        for el in els:
            linkID = el.get("id")
            if linkID is None:
                linkID = h.uniqueID("external-link", ref.url, termID) + str(counter)
                counter += 1
                el.set("id", h.safeID(doc, linkID))
            refsJson.append(
                {
                    "id": linkID,
                },
            )
        sectionsJson.append(
            {
                "refs": refsJson,
                "title": text,
            },
        )
    panelsJSON[termID] = {
        "dfnID": termID,
        "url": ref.url,
        "dfnText": termText,
        "refSections": sectionsJson,
        "external": True,
    }



from __future__ import annotations

from .. import t
from . import oldShorthands


def run(doc: t.SpecT) -> None:
    oldShorthands.transformMarkdownIB(doc)
    oldShorthands.transformProductionGrammars(doc)



from __future__ import annotations

import dataclasses
import json
import re
from abc import ABCMeta, abstractmethod
from pathlib import Path

from .. import config, h, t
from .. import messages as m


@dataclasses.dataclass
class JCManager:
    styles: dict[str, Style] = dataclasses.field(default_factory=dict)
    scripts: dict[str, Script] = dataclasses.field(default_factory=dict)

    def getStyles(self, allowList: t.BoolSet) -> list[Style]:
        styles: dict[str, Style] = {}
        for style in self.styles.values():
            if not style.insertable(allowList):
                continue
            styles[style.name] = style
        for script in self.scripts.values():
            if not script.insertable(allowList):
                continue
            if script.style and script.style.insertable(allowList):
                styles[script.style.name] = script.style
        return sorted(styles.values(), key=lambda x: x.name)

    def getScripts(self, allowList: t.BoolSet) -> list[Library | Script]:
        libs: dict[str, Library] = {}
        scripts: dict[str, Script] = {}
        for style in self.styles.values():
            if not style.insertable(allowList):
                continue
            if style.script and style.script.insertable(allowList):
                scripts[style.script.name] = style.script
        for script in self.scripts.values():
            if not script.insertable(allowList):
                continue
            if script.libraries:
                for lib in script.libraries.values():
                    if lib.insertable(allowList):
                        libs[lib.name] = lib
            scripts[script.name] = script
        return sorted(libs.values(), key=lambda x: x.name) + sorted(scripts.values(), key=lambda x: x.name)

    def _addCSS(self, name: str, moduleName: str = "stylescript") -> Style:
        if name not in self.styles:
            self.styles[name] = Style(name, Path(config.scriptPath(f"{moduleName}/{name}.css")))
        return self.styles[name]

    def _addJS(self, name: str, moduleName: str = "stylescript") -> Script:
        if name not in self.scripts:
            self.scripts[name] = Script(name, Path(config.scriptPath(f"{moduleName}/{name}.js")))
        return self.scripts[name]

    def addColors(self) -> None:
        self._addCSS("colors")

    def addMdLists(self) -> None:
        self._addCSS("md-lists")

    def addAutolinks(self) -> None:
        self._addCSS("autolinks")

    def addSelflinks(self) -> None:
        self._addCSS("selflinks")

    def addCounters(self) -> None:
        self._addCSS("counters")

    def addIssues(self) -> None:
        self._addCSS("issues")

    def addMdn(self) -> None:
        style = self._addCSS("mdn-anno", "mdn")
        if not style.script:
            style.script = Script("position-annos", Path(config.scriptPath("stylescript/position-annos.js")))

    def addCiu(self) -> None:
        style = self._addCSS("caniuse-panel", "caniuse")
        if not style.script:
            style.script = Script("position-annos", Path(config.scriptPath("stylescript/position-annos.js")))

    def addDomintro(self) -> None:
        self._addCSS("domintro")

    def addSyntaxHighlighting(self) -> None:
        self._addCSS("syntax-highlighting", "highlight")

    def addLineNumbers(self) -> None:
        self._addCSS("line-numbers", "highlight")

    def addLineHighlighting(self) -> None:
        self._addCSS("line-highlighting", "highlight")

    def addExpires(self) -> None:
        self._addJS("expires")

    def addHidedel(self) -> None:
        self._addCSS("hidedel")

    def addVarClickHighlighting(self) -> None:
        script = self._addJS("var-click-highlighting")
        if not script.style:
            script.style = Style(
                "var-click-highlighting",
                Path(config.scriptPath("stylescript/var-click-highlighting.css")),
            )

    def addRefHints(self) -> dict[str, t.Any]:
        script = self._addJS("ref-hints", "refs")
        if not script.style:
            script.style = Style("ref-hints", Path(config.scriptPath("refs/ref-hints.css")))
        if "dom-helper" not in script.libraries:
            script.libraries["dom-helper"] = Library("dom-helper", Path(config.scriptPath("stylescript/dom-helper.js")))
        if not script.data:
            script.data = ("refsData", {})
        return script.data[1]

    def addLinkTitles(self) -> dict[str, t.Any]:
        script = self._addJS("link-titles")
        if not script.data:
            script.data = ("linkTitleData", {})
        return script.data[1]

    def addCDDLHighlighting(self) -> None:
        self._addCSS("cddl-highlighting")

    def addIDLHighlighting(self) -> None:
        self._addCSS("idl-highlighting")

    def addRailroad(self) -> None:
        self._addCSS("railroad")

    def addWptCSS(self) -> None:
        self._addCSS("wpt", "wpt")

    def addWpt(self, paths: list[str] | None) -> None:
        script = self._addJS("wpt", "wpt")
        if not script.style:
            script.style = Style("wpt", Path(config.scriptPath("wpt/wpt.css")))
        if "dom-helper" not in script.libraries:
            script.libraries["dom-helper"] = Library("dom-helper", Path(config.scriptPath("stylescript/dom-helper.js")))
        if not script.data:
            script.data = ("wptData", {"paths": []})
        if paths:
            script.data[1]["paths"] = sorted(set(script.data[1]["paths"] + paths))

    def addDfnPanels(self) -> dict[str, t.Any]:
        script = self._addJS("dfn-panel", "dfnpanels")
        if not script.style:
            script.style = Style("dfn-panel", Path(config.scriptPath("dfnpanels/dfn-panel.css")))
        if "dom-helper" not in script.libraries:
            script.libraries["dom-helper"] = Library("dom-helper", Path(config.scriptPath("stylescript/dom-helper.js")))
        if not script.data:
            script.data = ("dfnPanelData", {})
        return script.data[1]


@dataclasses.dataclass
class JCResource(metaclass=ABCMeta):
    name: str

    @abstractmethod
    def insertable(self, allowList: t.BoolSet) -> bool:
        pass

    @abstractmethod
    def toElement(self) -> t.ElementT:
        pass


@dataclasses.dataclass
class Script(JCResource):
    path: Path
    libraries: dict[str, Library] = dataclasses.field(default_factory=dict)
    style: Style | None = None
    data: tuple[str, dict[str, t.Any]] | None = None

    def insertable(self, allowList: t.BoolSet) -> bool:
        if f"script-{self.name}" not in allowList:
            return False
        if self.data is None:
            return True
        if self.data[1]:  # noqa: SIM103
            return True
        return False

    def toElement(self) -> t.ElementT:
        text = f'/* Boilerplate: script-{self.name} */\n"use strict";\n{{\n'
        if self.data:
            text += f"let {self.data[0]} = {{\n"
            for key, val in sorted(self.data[1].items()):
                try:
                    jsonVal = val.__json__()
                except:
                    jsonVal = val
                text += f'"{key}": {json.dumps(jsonVal, sort_keys=True, separators=(",",":"))},\n'
            text += "};\n\n"
        with self.path.open("r", encoding="utf-8") as fh:
            text += fh.read()
        if not text.endswith("\n"):
            text += "\n"
        text += "}\n"
        return h.E.script(text)


@dataclasses.dataclass
class Library(JCResource):
    path: Path

    def insertable(self, allowList: t.BoolSet) -> bool:
        return True

    def toElement(self) -> t.ElementT:
        with self.path.open("r", encoding="utf-8") as fh:
            text = fh.read()
        return h.E.script(f'/* Boilerplate: script-{self.name} */\n"use strict";\n{text}')


@dataclasses.dataclass
class Style(JCResource):
    textPath: Path
    script: Script | None = None

    def insertable(self, allowList: t.BoolSet) -> bool:
        return f"style-{self.name}" in allowList

    def toElement(self, darkMode: bool = True) -> t.ElementT:
        with self.textPath.open("r", encoding="utf-8") as fh:
            sheet = fh.read()
        if not darkMode:
            sheet = removeInlineDarkStyles(self.name, sheet)
        if not sheet.endswith("\n"):
            sheet += "\n"
        return h.E.style(f"/* Boilerplate: style-{self.name} */\n{sheet}")


darkModeRe = re.compile(
    r"""
    ^@media\s+\(prefers-color-scheme:\s*dark\s*\)\s*{$ # start of block
    .*? # inside the block
    ^}(\n?)
    """,
    re.MULTILINE | re.DOTALL | re.X,
)
darkModeMaybe = re.compile(r"prefers-color-scheme\s*:\s*dark")


def removeInlineDarkStyles(name: str, text: str) -> str:
    # Looks for an @media declaring dark-mode styles,
    # and removes it from the text.
    # Simplistic: requires the @media and the closing }
    # to be against the left margin, and the block's contents
    # to all be indented.

    maybeMatch = re.search(darkModeMaybe, text)
    if not maybeMatch:
        return text
    match = re.search(darkModeRe, text)
    if not match:
        m.warn(
            f"The {name} stylesheet appears to contain darkmode styles, but they aren't being correctly detected. Please report this to the Bikeshed maintainer.",
        )
        return text
    text = re.sub(darkModeRe, "", text)
    # In case there are multiple blocks
    text = removeInlineDarkStyles(name, text)
    return text



from __future__ import annotations

import os

import requests

from .. import messages as m


def update(path: str, dryRun: bool = False) -> set[str] | None:
    try:
        m.say("Downloading link defaults...")
        data = requests.get(
            "https://raw.githubusercontent.com/speced/bikeshed/main/bikeshed/spec-data/readonly/link-defaults.infotree",
            timeout=5,
        ).text
    except Exception as e:
        m.die(f"Couldn't download link defaults data.\n{e}")
        return None

    filePath = os.path.join(path, "link-defaults.infotree")

    if not dryRun:
        try:
            with open(filePath, "w", encoding="utf-8") as f:
                f.write(data)
        except Exception as e:
            m.die(f"Couldn't save link-defaults database to disk.\n{e}")
            return None
    m.say("Success!")
    return {filePath}



from __future__ import annotations

import collections.abc

from .. import t


class BoolSet(collections.abc.MutableMapping):
    """
    Implements a "boolean set",
    where keys can be explicitly set to True or False,
    but interacted with like a normal set
    (similar to Counter, but with bools).
    Can also set whether the default should consider unset values to be True or False by default.
    """

    def __init__(self, values: t.Any = None, default: bool = False) -> None:
        self._internal: dict[t.Any, bool] = {}
        if isinstance(values, collections.abc.Mapping):
            for k, v in values.items():
                self._internal[k] = bool(v)
        elif isinstance(values, collections.abc.Iterable):
            for k in values:
                self._internal[k] = True
        self.default = bool(default)

    def __missing__(self, key: t.Any) -> bool:
        return self.default

    def __contains__(self, key: t.Any) -> bool:
        if key in self._internal:
            return self._internal[key]
        else:
            return self.default

    def __getitem__(self, key: t.Any) -> bool:
        return key in self

    def __setitem__(self, key: t.Any, val: bool) -> None:
        self._internal[key] = bool(val)

    def __delitem__(self, key: t.Any) -> None:
        del self._internal[key]

    def __iter__(self) -> t.Iterator[t.Any]:
        return iter(self._internal)

    def __len__(self) -> int:
        return len(self._internal)

    def __repr__(self) -> str:
        if self.default is False:
            trueVals = [k for k, v in self._internal.items() if v is True]
            vrepr = "[" + ", ".join(repr(x) for x in trueVals) + "]"
        else:
            falseVals = [k for k, v in self._internal.items() if v is False]
            vrepr = "{" + ", ".join(repr(x) + ":False" for x in falseVals) + "}"
        return f"BoolSet({vrepr}, default={self.default})"

    def hasExplicit(self, key: t.Any) -> bool:
        return key in self._internal

    @t.overload
    def update(self, __other: t.SupportsKeysAndGetItem[t.Any, t.Any], **kwargs: bool) -> None: ...

    @t.overload
    def update(self, __other: t.Iterable[tuple[t.Any, t.Any]], **kwargs: bool) -> None: ...

    @t.overload
    def update(self, **kwargs: bool) -> None: ...

    def update(self, __other: t.Any = None, **kwargs: bool) -> None:
        # If either defaults to True, the result
        # should too, since it "contains" everything
        # by default.
        if isinstance(__other, BoolSet):
            self.default = self.default or __other.default
            self._internal.update(__other._internal)  # pylint: disable=protected-access
        elif __other is not None:
            self._internal.update(__other)
        for k, v in kwargs.items():
            self[k] = bool(v)



from __future__ import annotations

from . import t


class Functor:
    # Pointed and Co-Pointed by default;
    # override these yourself if you need to.
    def __init__(self, val: t.Any) -> None:
        self.__val__ = val

    def extract(self) -> t.Any:
        return self.__val__

    def map(self, fn: t.Any) -> t.Any:
        return self.__class__(fn(self.__val__))



from __future__ import annotations

import dataclasses
import re

from . import h, t
from . import messages as m

# Any element can have an include-if or exclude-if attribute,
# containing a comma-separated list of conditions (described below).
# If an element has include-if, it must match at least one condition,
# or else it's removed from the document.
# If an element has exclude-if, if it matches at least one condition,
# it's removed from the document.
# (An element can have both; both of the above conditions apply.)
#
# A condition is either a Status value, or a `!Foo: bar` custom metadata declaration,
# which is compared literally after trimming.
#
# The <if-wrapper> element is a special wrapper element
# that exists *solely* to support conditional inclusion,
# and which never appears in the output document.
# It must have a conditional attribute on it,
# and if it passes the checks,
# it's still removed from the document,
# but its children are left in its place.


def processConditionals(doc: t.SpecT, container: t.ElementT | None = None) -> None:
    for el in h.findAll("[include-if], [exclude-if], if-wrapper", container if container is not None else doc):
        if el.tag == "if-wrapper" and not h.hasAnyAttr(el, "include-if", "exclude-if"):
            m.die(
                "<if-wrapper> elements must have an include-if and/or exclude-if attribute.",
                el=el,
            )
            h.removeNode(el)
            continue

        removeEl = False
        if h.hasAttr(el, "include-if"):
            if not any(evalConditions(doc, el, el.get("include-if", ""))):
                removeEl = True
        if not removeEl and h.hasAttr(el, "exclude-if"):
            if any(evalConditions(doc, el, el.get("exclude-if", ""))):
                removeEl = True

        if removeEl:
            h.removeNode(el)
            continue

        if el.tag == "if-wrapper":
            # Passed the tests, so just remove the wrapper
            h.replaceNode(el, *h.childNodes(el, clear=True))
            continue

        # Otherwise, go ahead and just remove the include/exclude-if attributes,
        # since they're non-standard
        h.removeAttr(el, "include-if", "exclude-if")


def evalConditions(doc: t.SpecT, el: t.ElementT, conditionString: str) -> t.Generator[bool, None, None]:
    for cond in parseConditions(conditionString, el):
        if cond.type == "status":
            yield doc.doctype.status.looselyMatch(cond.value)
        elif cond.type == "text macro":
            for k in doc.macros:
                if k.upper() == cond.value:
                    yield True
                    break
            else:
                yield False
        elif cond.type == "boilerplate":
            yield (h.find(f'[boilerplate="{h.escapeCSSIdent(cond.value)}"]', doc) is not None)
        else:
            m.die(
                f"Program error, some type of include/exclude-if condition wasn't handled: '{cond!r}'. Please report!",
                el,
            )
            yield False


@dataclasses.dataclass
class Condition:
    type: str
    value: str


def parseConditions(s: str, el: t.ElementT | None = None) -> t.Generator[Condition, None, None]:
    for sub in s.split(","):
        sub = sub.strip()
        if sub == "":
            continue
        if re.match(r"([\w-]+/)?[\w-]+$", sub):
            yield Condition(type="status", value=sub)
            continue
        match = re.match(r"text macro:\s*(.+)$", sub, re.I)
        if match:
            yield Condition(type="text macro", value=match.group(1).strip())
            continue
        match = re.match(r"boilerplate:\s*(.+)$", sub, re.I)
        if match:
            yield Condition(type="boilerplate", value=match.group(1).strip())
            continue
        m.die(f"Unknown include/exclude-if condition '{sub}'", el=el)
        continue



from __future__ import annotations

import dataclasses
import re
from abc import ABCMeta, abstractmethod
from collections import Counter
from dataclasses import InitVar, dataclass, field

from ... import messages as m
from ... import t

if t.TYPE_CHECKING:
    from .stream import Stream  # pylint: disable=cyclic-import


@dataclass
class ParserNode(metaclass=ABCMeta):
    line: int
    endLine: int
    loc: str
    endLoc: str
    context: str | None

    @property
    def height(self) -> int:
        return self.endLine - self.line


@dataclass
class Text(ParserNode, metaclass=ABCMeta):
    text: str

    @abstractmethod
    def __str__(self) -> str:
        pass


@dataclass
class RawText(Text):
    # Raw source text, might contain HTML characters/etc

    def __str__(self) -> str:
        return self.text

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, text: str | None = None) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            text=text if text is not None else s.slice(start, end),
        )

    @classmethod
    def fromSafeText(cls, node: SafeText) -> t.Self:
        return cls(
            line=node.line,
            endLine=node.endLine,
            loc=node.loc,
            endLoc=node.endLoc,
            context=node.context,
            text=str(node),
        )

    def curlifyApostrophes(self, lastNode: ParserNode | None) -> RawText:
        if re.match(r"'\w", self.text):
            if isinstance(lastNode, (EndTag, RawElement, SelfClosedTag)):
                self.text = "’" + self.text[1:]
            elif isinstance(lastNode, RawText) and re.match(r"\w", lastNode.text[-1]):
                self.text = "’" + self.text[1:]
        if "'" in self.text:
            self.text = re.sub(r"(\w)'(\w)", r"\1’\2", self.text)
        if isinstance(lastNode, EndTag) and lastNode.tag == "var" and re.match(r"'\s", self.text):
            self.text = "’" + self.text[1:]
        return self

    def needsLCCs(self) -> bool:
        """
        Whether or not the node will eventally insert an ILCC or DLCC
        to fix the line count when serializing to a string.
        """
        return self.text.count("\n") != self.height


@dataclass
class SafeText(Text):
    # "Safe" text, automatically escapes special HTML chars
    # when stringified.
    def __str__(self) -> str:
        return escapeHTML(self.text)

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, text: str | None = None) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            text=text if text is not None else s.slice(start, end),
        )


@dataclass
class Doctype(ParserNode):
    data: str

    def __str__(self) -> str:
        return self.data

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, data: str) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            data=data,
        )


@dataclass
class StartTag(ParserNode):
    tag: str
    attrs: dict[str, str] = field(default_factory=dict)
    classes: set[str] = field(default_factory=set)

    @classmethod
    def fromStream(
        cls,
        s: Stream,
        start: int,
        end: int,
        tag: str,
        attrs: None | dict[str, str] = None,
    ) -> t.Self:
        if attrs is None:
            attrs = {}
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            tag=tag,
            attrs=attrs,
        )

    def __str__(self) -> str:
        s = f'<{self.tag} bs-line-number="{escapeAttr(self.loc)}"'
        if self.context:
            s += f' bs-parse-context="{escapeAttr(self.context)}"'
        for k, v in sorted(self.attrs.items()):
            if k == "bs-line-number":
                continue
            v = v.replace('"', "&#34;")
            s += f' {k}="{v}"'
        if self.classes:
            s += f' class="{" ".join(sorted(self.classes))}"'
        s += ">"
        return s

    def printEndTag(self) -> str:
        return f"</{self.tag}>"

    def finalize(self) -> StartTag:
        if "class" in self.attrs:
            self.classes = set(self.attrs["class"].split())
            del self.attrs["class"]
        return self

    def clone(self, **kwargs: t.Any) -> StartTag:
        return dataclasses.replace(self, **kwargs)


@dataclass
class SelfClosedTag(ParserNode):
    tag: str
    attrs: dict[str, str] = field(default_factory=dict)
    classes: set[str] = field(default_factory=set)

    @classmethod
    def fromStream(
        cls,
        s: Stream,
        start: int,
        end: int,
        tag: str,
        attrs: None | dict[str, str] = None,
    ) -> t.Self:
        if attrs is None:
            attrs = {}
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            tag=tag,
            attrs=attrs,
        )

    @classmethod
    def fromStartTag(cls, tag: StartTag) -> t.Self:
        return cls(
            line=tag.line,
            endLine=tag.endLine,
            loc=tag.loc,
            endLoc=tag.endLoc,
            context=tag.context,
            tag=tag.tag,
            attrs=tag.attrs,
            classes=tag.classes,
        )

    def __str__(self) -> str:
        s = f'<{self.tag} bs-line-number="{escapeAttr(self.loc)}"'
        if self.context:
            s += f' bs-parse-context="{escapeAttr(self.context)}"'
        for k, v in sorted(self.attrs.items()):
            if k == "bs-line-number":
                continue
            v = v.replace('"', "&#34;")
            s += f' {k}="{v}"'
        if self.classes:
            s += f' class="{" ".join(sorted(self.classes))}"'
        if self.tag in (
            "area",
            "base",
            "br",
            "col",
            "command",
            "embed",
            "hr",
            "img",
            "input",
            "keygen",
            "link",
            "meta",
            "param",
            "source",
            "track",
            "wbr",
        ):
            s += ">"
        else:
            s += f"></{self.tag}>"
        return s

    def finalize(self) -> SelfClosedTag:
        if "class" in self.attrs:
            self.classes = set(self.attrs["class"].split())
            del self.attrs["class"]
        return self

    def clone(self, **kwargs: t.Any) -> SelfClosedTag:
        return dataclasses.replace(self, **kwargs)


@dataclass
class EndTag(ParserNode):
    tag: str

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, tag: str | StartTag) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            tag=tag if isinstance(tag, str) else tag.tag,
        )

    def __str__(self) -> str:
        return f"</{self.tag}>"


@dataclass
class Comment(ParserNode):
    data: str

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, data: str) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            data=data,
        )

    def __str__(self) -> str:
        return f"<!--{escapeHTML(self.data)}-->"


# RawElement is for things like <script> or <xmp>
# which have special parsing rules that just look
# for the ending tag and treat the entire rest of
# the contents as raw text, without escaping.
@dataclass
class RawElement(ParserNode):
    tag: str
    startTag: StartTag
    data: str

    @classmethod
    def fromStream(cls, s: Stream, start: int, end: int, startTag: StartTag, data: str) -> t.Self:
        return cls(
            line=s.line(start),
            endLine=s.line(end),
            loc=s.loc(start),
            endLoc=s.loc(end),
            context=s.context,
            tag=startTag.tag,
            startTag=startTag,
            data=data,
        )

    def __str__(self) -> str:
        return f"{self.startTag}{self.data}</{self.tag}>"


def escapeHTML(text: str) -> str:
    # Escape HTML
    return text.replace("&", "&amp;").replace("<", "&lt;")


def escapeAttr(text: str) -> str:
    return text.replace("&", "&amp;").replace('"', "&quot;")


@dataclass
class TagStack:
    tags: list[TagStackEntry] = field(default_factory=list)
    opaqueTags: set[str] = field(default_factory=lambda: {"pre", "xmp", "script", "style"})
    _opaqueCount: int = 0
    _tagCounts: Counter[str] = field(default_factory=Counter)

    def printOpenTags(self) -> list[str]:
        return [f"{x.name} at {x.startTag.loc}" for x in self.tags]

    def inOpaqueElement(self) -> bool:
        return self._opaqueCount > 0

    def inTagContext(self, tagName: str) -> bool:
        return self._tagCounts[tagName] > 0

    def pushEntry(self, entry: TagStackEntry) -> None:
        self.tags.append(entry)
        if entry.isOpaque:
            self._opaqueCount += 1
        self._tagCounts[entry.startTag.tag] += 1

    def popEntry(self) -> None:
        entry = self.tags.pop()
        if entry.isOpaque:
            self._opaqueCount -= 1
        self._tagCounts[entry.startTag.tag] -= 1

    def update(self, node: ParserNode) -> None:
        # Updates the stack based on the passed node.
        # Start tags add to the stack, close tags pop from it.
        # Auto-closing tags mean start tags can also pop from the stack.
        if isinstance(node, StartTag):
            self.autoCloseStart(node.tag)
            self.pushEntry(TagStackEntry(node, self.opaqueTags))
            return
        elif isinstance(node, EndTag):
            self.autoCloseEnd(node.tag)
            if (
                self.tags
                and self.tags[-1].startTag.tag == node.tag
                and not isinstance(self.tags[-1], TagStackShorthandEntry)
            ):
                self.popEntry()
            else:
                if node.tag in ("html", "head", "body", "main"):
                    # If your boilerplate closes these for safety,
                    # but they're not actually open,
                    # that's fine
                    return
                for entry in reversed(self.tags):
                    if entry.startTag.tag == node.tag:
                        m.die(
                            f"Saw an end tag {node}, but there were unclosed elements remaining before the nearest matching start tag (at {entry.startTag.loc}).\nOpen tags: {', '.join(self.printOpenTags())}",
                            lineNum=node.loc,
                        )
                        break
                else:
                    openTagsMsg = f"\nOpen tags: {', '.join(self.printOpenTags())}" if self.tags else ""
                    m.die(
                        f"Saw an end tag {node}, but there's no open element corresponding to it.{openTagsMsg}",
                        lineNum=node.loc,
                    )
        elif isinstance(node, (SelfClosedTag, RawElement)):
            self.autoCloseStart(node.tag)

    def updateShorthandOpen(self, startTag: StartTag, sigils: tuple[str, str]) -> None:
        self.pushEntry(TagStackShorthandEntry(startTag, self.opaqueTags, sigils))

    def updateShorthandClose(self, loc: str, startTag: StartTag, sigils: tuple[str, str]) -> None:
        if self.tags and self.tags[-1].startTag == startTag:
            self.popEntry()
            return
        shorthand = sigils[0] + "..." + sigils[1]
        if any(x.startTag == startTag for x in self.tags):
            m.die(
                f"{shorthand} shorthand (opened on {startTag.loc}) was closed, but there were still open elements inside of it.\nOpen tags: {', '.join(self.printOpenTags())}",
                lineNum=loc,
            )
            while self.tags and self.tags[-1].startTag != startTag:
                self.popEntry()
            if self.tags:
                self.popEntry()
        else:
            m.die(
                f"PROGRAMMING ERROR: Tried to close a {shorthand} shorthand, but there's no matching open tag on the stack of open elements. Please report this!",
                lineNum=loc,
            )

    def cancelShorthandOpen(self, startTag: StartTag, sigils: tuple[str, str]) -> None:
        if not any(x.startTag == startTag for x in self.tags):
            shorthand = sigils[0] + "..." + sigils[1]
            m.die(
                f"Programming error - tried to close a {shorthand} shorthand, but there's no matching open tag on the stack of open elements. Please report this!",
                lineNum=startTag.loc,
            )
            return
        while self.tags and self.tags[-1].startTag != startTag:
            self.popEntry()
        self.popEntry()

    def autoCloseStart(self, tag: str) -> None:
        # Handle any auto-closing that occurs as a result
        # of seeing a particular start tag.
        # (Only valid HTML; real parsers do a lot more work.)
        if not self.tags:
            return
        if tag in (
            "address",
            "article",
            "aside",
            "blockquote",
            "center",
            "details",
            "dialog",
            "dir",
            "div",
            "dl",
            "fieldset",
            "figcaption",
            "figure",
            "footer",
            "header",
            "hgroup",
            "main",
            "menu",
            "nav",
            "ol",
            "p",
            "search",
            "section",
            "summary",
            "ul",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "pre",
            "listing",
            "form",
            "plaintext",
            "table",
            "hr",
            "xmp",
        ):
            self.virtualClose("p")
        if tag == "li":
            self.virtualClose("p")
            self.virtualClose("li")
        elif tag in ("dt", "dd"):
            self.virtualClose("p")
            self.virtualClose("dt", "dd")
        elif tag == "option":
            self.virtualClose("option")
        elif tag == "optgroup":
            self.virtualClose("option")
            self.virtualClose("optgroup")
        elif tag in ("td", "th"):
            self.virtualClose("p")
            self.virtualClose("td", "th")
        elif tag == "tr":
            self.virtualClose("p")
            self.virtualClose("td", "th")
            self.virtualClose("tr")
            self.virtualClose("caption", "colgroup")
        elif tag in ("thead", "tbody", "tfoot"):
            self.virtualClose("p")
            self.virtualClose("td", "th")
            self.virtualClose("tr")
            self.virtualClose("thead", "tbody", "tfoot", "caption", "colgroup")
        elif tag in ("caption", "colgroup"):
            self.virtualClose("p")
            self.virtualClose("td", "th")
            self.virtualClose("tr")
            self.virtualClose("thead", "tbody", "tfoot", "caption", "colgroup")
        elif tag == "col":
            self.virtualClose("p")
            self.virtualClose("td", "th")
            self.virtualClose("tr")
            self.virtualClose("thead", "tbody", "tfoot", "caption")
        elif tag in ("rb", "rtc"):
            self.virtualClose("rb", "rp", "rt")
            self.virtualClose("rtc")
        elif tag in ("rp", "rt"):
            self.virtualClose("rb", "rp", "rt")

    def autoCloseEnd(self, tag: str) -> None:
        # Handle any auto-closing that occurs as a result
        # of seeing a particular end tag.
        # (Only valid HTML; real parsers do a lot more work.)
        if tag in (
            "address",
            "article",
            "aside",
            "blockquote",
            "center",
            "details",
            "dialog",
            "dir",
            "div",
            "dl",
            "fieldset",
            "figcaption",
            "figure",
            "footer",
            "header",
            "hgroup",
            "main",
            "menu",
            "nav",
            "ol",
            "search",
            "section",
            "summary",
            "ul",
            "h1",
            "h2",
            "h3",
            "h4",
            "h5",
            "h6",
            "pre",
            "listing",
            "form",
            "plaintext",
            "table",
            "caption",
            "thead",
            "tbody",
            "tfoot",
            "tr",
            "td",
            "th",
            "li",
            "dt",
            "dd",
        ):
            self.virtualClose("p")
        if tag in ("ol", "ul"):
            self.virtualClose("li")
        elif tag == "dl":
            self.virtualClose("dt", "dd")
        elif tag == "tr":
            self.virtualClose("td", "th")
        elif tag in ("thead", "tbody", "tfoot"):
            self.virtualClose("td", "th")
            self.virtualClose("tr")
        elif tag == "table":
            self.virtualClose("td", "th")
            self.virtualClose("tr")
            self.virtualClose("thead", "tbody", "tfoot", "caption", "colgroup")
        elif tag == "rtc":
            self.virtualClose("rt")
        elif tag == "ruby":
            self.virtualClose("rt", "rb", "rp")
            self.virtualClose("rtc")
        elif tag == "optgroup":
            self.virtualClose("option")
        elif tag == "select":
            self.virtualClose("option")
            self.virtualClose("optgroup")

    def virtualClose(self, *tags: str) -> None:
        if not self.tags:
            return
        if self.tags[-1].startTag.tag in tags:
            self.popEntry()


@dataclass
class TagStackEntry:
    startTag: StartTag
    opaqueTags: InitVar[set[str]]
    isOpaque: bool = field(init=False)

    def __post_init__(self, opaqueTags: set[str]) -> None:
        if self.startTag.tag in opaqueTags or "bs-opaque" in self.startTag.attrs:
            self.isOpaque = True
        else:
            self.isOpaque = False

    @property
    def name(self) -> str:
        return "<" + self.startTag.tag + ">"


@dataclass
class TagStackShorthandEntry(TagStackEntry):
    sigils: tuple[str, str]

    @property
    def name(self) -> str:
        return self.sigils[0] + "..." + self.sigils[1]



from __future__ import annotations

import json
import os
from collections import OrderedDict

import requests

from .. import messages as m


def update(path: str, dryRun: bool = False) -> set[str] | None:
    m.say("Downloading MDN Spec Links data...")
    specMapURL = "https://w3c.github.io/mdn-spec-links/SPECMAP.json"
    try:
        response = requests.get(specMapURL, timeout=5)
    except Exception as e:
        m.die(f"Couldn't download the MDN Spec Links data.\n{e}")
        return None

    try:
        data = response.json(object_pairs_hook=OrderedDict)
    except Exception as e:
        m.die(f"The MDN Spec Links data wasn't valid JSON for some reason. Try downloading again?\n{e}")
        return None
    writtenPaths = set()
    if not dryRun:
        try:
            mdnSpecLinksDir = os.path.join(path, "mdn")
            if not os.path.exists(mdnSpecLinksDir):
                os.makedirs(mdnSpecLinksDir)
            p = os.path.join(path, "mdn.json")
            writtenPaths.add(p)
            with open(p, "w", encoding="utf-8") as fh:
                fh.write(json.dumps(data, indent=1, ensure_ascii=False, sort_keys=False))
            # SPECMAP.json format:
            # {
            #     "https://compat.spec.whatwg.org/": "compat.json",
            #     "https://console.spec.whatwg.org/": "console.json",
            #     "https://dom.spec.whatwg.org/": "dom.json",
            #     ...
            # }
            for specFilename in data.values():
                p = os.path.join(mdnSpecLinksDir, specFilename)
                writtenPaths.add(p)
                mdnSpecLinksBaseURL = "https://w3c.github.io/mdn-spec-links/"
                try:
                    fileContents = requests.get(mdnSpecLinksBaseURL + specFilename, timeout=5).text
                except Exception as e:
                    m.die(f"Couldn't download the MDN Spec Links {specFilename} file.\n{e}")
                    return None
                with open(p, "w", encoding="utf-8") as fh:
                    fh.write(fileContents)
        except Exception as e:
            m.die(f"Couldn't save MDN Spec Links data to disk.\n{e}")
            return None
    m.say("Success!")
    return writtenPaths



from __future__ import annotations

import re

from .. import h, t
from .. import messages as m


def accidental2119(doc: t.SpecT) -> None:
    """
    Looks for usage of 2119 keywords in non-normative sections.
    You can override this, allowing the keyword,
    by putting an "allow-2119" class on the text's containing element specifically
    (not an ancestor, to avoid accidentally over-silencing).
    """
    if not doc.md.complainAbout["accidental-2119"]:
        return
    keywords = r"\b(may|must|should|shall|optional|recommended|required)\b"

    def searchFor2119(el: t.ElementT) -> None:
        if h.isNormative(doc, el):
            # 2119 is fine, just look at children
            pass
        elif h.hasClass(doc, el, "allow-2119"):
            # Override 2119 detection on this element's text specifically,
            # so you can use the keywords in examples *describing* the keywords.
            pass
        else:
            if el.text is not None:
                match = re.search(keywords, el.text)
                if match:
                    m.lint(
                        f"RFC2119 keyword in non-normative section (use: might, can, has to, or override with <span class=allow-2119>): {el.text}",
                        el=el,
                    )
            for child in el:
                if child.tail is not None:
                    match = re.search(keywords, child.tail)
                    if match:
                        m.lint(
                            f"RFC2119 keyword in non-normative section (use: might, can, has to, or override with <span class=allow-2119>): {child.tail}",
                            el=el,
                        )
        for child in el:
            searchFor2119(child)

    searchFor2119(doc.body)



from __future__ import annotations

import asyncio
import dataclasses
import hashlib
import os
import sys
import time
from datetime import datetime, timezone

import aiofiles
import aiohttp
import kdl
import requests
import tenacity
from result import Err, Ok, Result

from .. import messages as m
from .. import t
from .mode import UpdateMode


def isOk(x: t.Any) -> t.TypeGuard[Ok]:
    return isinstance(x, Ok)


def isErr(x: t.Any) -> t.TypeGuard[Err]:
    return isinstance(x, Err)


# Manifest creation relies on these data structures.
# Add to them whenever new types of data files are created.
knownFiles = [
    "biblio-keys.json",
    "biblio-numeric-suffixes.json",
    "fors.json",
    "languages.json",
    "link-defaults.infotree",
    "mdn.json",
    "methods.json",
    "specs.json",
    "wpt-tests.txt",
]
knownFolders = [
    "anchors",
    "biblio",
    "boilerplate",
    "caniuse",
    "headings",
    "mdn",
]

ghPrefix = "https://raw.githubusercontent.com/speced/bikeshed-data/main/data/"

# To avoid 'Event loop is closed' RuntimeError due to compatibility issue with aiohttp
if sys.platform.startswith("win") and sys.version_info >= (3, 8):
    try:
        from asyncio import WindowsSelectorEventLoopPolicy
    except ImportError:
        pass
    else:
        if not isinstance(asyncio.get_event_loop_policy(), WindowsSelectorEventLoopPolicy):
            asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


def createManifest(path: str, dryRun: bool = False) -> Manifest:
    """Generates a manifest file for all the data files."""
    manifest = Manifest.fromPath(path)
    if not dryRun:
        manifest.save(path)
    return manifest


def updateByManifest(path: str, dryRun: bool = False, updateMode: UpdateMode = UpdateMode.MANIFEST) -> Manifest | None:
    """
    Attempts to update only the recently updated datafiles by using a manifest file.
    Returns None if updating failed and a full update should be performed;
    returns the manifest if updating was a success.
    """
    m.say("Updating via manifest...")

    m.say("Gathering local manifest data...")
    # Get the last-update time from the local manifest
    try:
        with open(os.path.join(path, "manifest.txt"), encoding="utf-8") as fh:
            oldManifest = Manifest.fromString(fh.read())
    except Exception as e:
        oldManifest = None
        m.warn(f"Couldn't find manifest from previous update run:\n  {e}")

    # Get the actual file data by regenerating the local manifest,
    # to guard against mistakes or shenanigans
    localManifest = Manifest.fromPath(path)
    if oldManifest:
        localManifest.dt = oldManifest.dt
    else:
        # Ensure it thinks it's old
        localManifest.dt = dtZero()

    m.say("Fetching remote manifest data...")
    try:
        remoteManifestText = requests.get(ghPrefix + "manifest.txt", timeout=5).text
    except Exception as e:
        m.warn(
            f"Couldn't download remote manifest file, so can't update. Please report this!\n{e}",
        )
        m.warn("If absolutely necessary, you can update manually with `bikeshed update --skip-manifest`.")
        return None
    remoteManifest = Manifest.fromString(remoteManifestText)

    if remoteManifest is None:
        m.die("Something's gone wrong with the remote data; I can't read its timestamp. Please report this!")
        return None

    if oldManifest and oldManifest.hasError():
        # A previous update run didn't complete successfully,
        # so I definitely need to try again.

        m.warn("Previous update had some download errors, so re-running...")
    else:
        if remoteManifest.daysOld() > 2:
            m.warn(
                f"Remote data ({printDt(remoteManifest.dt)}) is more than two days older than local time ({printDt(dtNow())}); either your local time is wrong (no worries, this warning will just repeat each time) or the update process has fallen over (please report this!).",
            )
        if not (updateMode & UpdateMode.FORCE):
            # If the update isn't forced, allow it to be skipped
            # if the data is already sufficiently fresh.
            if localManifest.dt == remoteManifest.dt and localManifest.dt == dtZero():
                m.say(f"Local data is already up-to-date with remote ({printDt(localManifest.dt)})")
                return localManifest
            elif localManifest.dt > remoteManifest.dt:
                # No need to update, local data is more recent.
                m.say(
                    f"Local data is fresher ({printDt(localManifest.dt)}) than remote ({printDt(remoteManifest.dt)}), so nothing to update.",
                )
                return localManifest

    if not oldManifest or len(oldManifest.entries) == 0:
        m.say("The local manifest is borked or missing; re-downloading everything...")
    if len(remoteManifest.entries) == 0:
        m.die("The remote data doesn't have any data in it. Please report this!")
        return None
    newPaths = []
    for filePath, hash in remoteManifest.entries.items():
        if hash != localManifest.entries.get(filePath):
            # print(f"{hash} {filePath} {localManifest.entries.get(filePath)}")
            newPaths.append(filePath)
    if not dryRun:
        deletedPaths = []
        for filePath in localManifest.entries:
            relPath = localizePath(path, filePath)
            if filePath not in remoteManifest.entries and os.path.exists(relPath):
                os.remove(relPath)
                deletedPaths.append(filePath)
        if deletedPaths:
            m.say("Deleted {} old data file{}.".format(len(deletedPaths), "s" if len(deletedPaths) > 1 else ""))

    newManifest = None
    if not dryRun:
        if newPaths:
            m.say(f"Updating {len(newPaths)} file{'s' if len(newPaths) > 1 else ''}...")
        _, badPaths = asyncio.run(updateFiles(path, newPaths))
        newManifest = dataclasses.replace(remoteManifest)
        for badPath in badPaths:
            newManifest.entries[badPath] = None
        try:
            with open(os.path.join(path, "manifest.txt"), "w", encoding="utf-8") as fh:
                fh.write(str(newManifest))
        except Exception as e:
            m.warn(f"Couldn't save new manifest file.\n{e}")
            return None
    if newManifest is None:
        newManifest = Manifest.fromPath(path)

    if not badPaths:
        m.say("Done!")
        return newManifest
    else:
        phrase = f"were {len(badPaths)} errors" if len(badPaths) > 1 else "was 1 error"
        m.die(
            f"Done, but there {phrase} (of {len(newPaths)} total) in downloading or saving. Run `bikeshed update` again to retry.",
        )
        return newManifest


async def updateFiles(localPrefix: str, newPaths: list[str]) -> tuple[list[str], list[str]]:
    tasks = set()
    async with aiohttp.ClientSession(trust_env=True) as session:
        for filePath in newPaths:
            coro = updateFile(localPrefix, filePath, session=session)
            tasks.add(coro)

        lastMsgTime = time.time()
        messageDelta = 2
        goodPaths = []
        badPaths = []
        for future in asyncio.as_completed(tasks):
            result = await future
            if isOk(result):
                goodPaths.append(result.value)
            else:
                badPaths.append(result.value)
            currFileTime = time.time()
            if (currFileTime - lastMsgTime) >= messageDelta:
                if not badPaths:
                    m.say(f"Updated {len(goodPaths)}/{len(newPaths)}...")
                else:
                    m.say(f"Updated {len(goodPaths)}/{len(newPaths)}, {len(badPaths)} errors...")
                lastMsgTime = currFileTime
    return goodPaths, badPaths


async def updateFile(localPrefix: str, filePath: str, session: t.Any) -> Result[str, str]:
    remotePath = ghPrefix + filePath
    localPath = localizePath(localPrefix, filePath)
    res = await downloadFile(remotePath, session)
    if isOk(res):
        res = await saveFile(localPath, res.ok())
    else:
        m.warn(f"Error downloading {filePath}, full error was:\n{await errorFromAsyncErr(res)}")
    ret: Result[str, str]
    if isErr(res):
        ret = Err(filePath)
    else:
        ret = t.cast("Ok[str]", res)
    return ret


async def errorFromAsyncErr(res: Result[str, t.Awaitable[str]]) -> str | Exception:
    if isOk(res):
        return t.cast(str, res.ok())
    try:
        x = await t.cast("t.Awaitable[str]", res.err())
    except Exception as e:
        return e
    return x


def wrapError(retry_state: t.Any) -> Err[t.Awaitable[str]]:
    return Err(asyncio.wrap_future(retry_state.outcome))


@tenacity.retry(
    reraise=True,
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_random(1, 2),
    retry_error_callback=wrapError,
)
async def downloadFile(path: str, session: t.Any) -> Result[str, t.Awaitable[str]]:
    resp = await session.request(method="GET", url=path)
    resp.raise_for_status()
    return Ok(await resp.text())


@tenacity.retry(
    reraise=True,
    stop=tenacity.stop_after_attempt(3),
    wait=tenacity.wait_random(1, 2),
    retry_error_callback=wrapError,
)
async def saveFile(path: str, data: str) -> Result[str, t.Awaitable[str]]:
    dirPath = os.path.dirname(path)
    if not os.path.exists(dirPath):
        os.makedirs(dirPath)
    async with aiofiles.open(path, "w", encoding="utf-8") as fh:
        await fh.write(data)
        return Ok(path)


def localizePath(root: str, relPath: str) -> str:
    return os.path.join(root, *relPath.split("/"))


if t.TYPE_CHECKING:
    ManifestRelPath: t.TypeAlias = str
    ManifestFileHash: t.TypeAlias = str | None


@dataclasses.dataclass
class Manifest:
    dt: datetime = dataclasses.field(default_factory=lambda: dtNow())
    entries: dict[ManifestRelPath, ManifestFileHash] = dataclasses.field(default_factory=dict)
    # Bump this version manually whenever you update the datafiles
    version: int = 1

    @staticmethod
    def fromString(text: str) -> Manifest | None:
        try:
            doc = kdl.parse(text)
        except kdl.errors.ParseError:
            return Manifest.legacyFromString(text)
        manifest = doc["manifest"]
        entries = {file.props["path"]: file.props["hash"] for file in manifest.getAll("file")}
        return Manifest(dt=manifest.props["updated"], version=manifest.props["version"], entries=entries)

    @staticmethod
    def legacyFromString(text: str) -> Manifest | None:
        lines = text.split("\n")
        if len(lines) < 10:
            # Something's definitely borked
            m.warn(
                f"Error when parsing manifest: manifest is {len(lines)} long, which is too short to possibly be valid. Please report this!\nEntire manifest:\n{text}",
            )
            return None
        dt = parseDt(lines[0].strip())
        if dt is None:
            return None
        try:
            version = int(lines[1].strip())
            entryLines = lines[2:]
        except ValueError:
            version = 0
            entryLines = lines[1:]
        entries: dict[str, str | None] = {}
        for line in entryLines:
            line = line.strip()
            if line == "":
                continue
            hash, _, path = line.strip().partition(" ")
            if hash.startswith("error"):
                entries[path.strip()] = None
            else:
                entries[path.strip()] = hash

        return Manifest(dt, entries, version)

    @staticmethod
    def fromPath(path: str) -> Manifest:
        manifest = Manifest()
        for absPath, relPath in getDatafilePaths(path):
            if relPath in knownFiles:
                pass
            elif relPath.partition("/")[0] in knownFolders:
                pass
            else:
                continue
            with open(absPath, encoding="utf-8") as fh:
                manifest.entries[relPath] = hashFile(fh)
        return manifest

    def __str__(self) -> str:
        doc = kdl.Document(
            nodes=[
                kdl.Node(
                    "manifest",
                    None,
                    props={"updated": self.dt, "version": self.version},
                ),
            ],
        )
        manifestNode = doc["manifest"]
        for p, h in sorted(self.entries.items(), key=keyManifest):
            node = kdl.Node("file", None, props={"hash": h, "path": p})
            if h is None:
                node.props["error"] = True
            manifestNode.nodes.append(node)
        return doc.print()

    def daysOld(self) -> int:
        return (dtNow() - self.dt).days

    def save(self, path: str) -> None:
        with open(os.path.join(path, "manifest.txt"), "w", encoding="utf-8") as fh:
            fh.write(str(self))

    def hasError(self) -> bool:
        return self.dt == dtZero() or any(x is None for x in self.entries.values())


def keyManifest(entry: tuple[str, t.Any]) -> tuple[int, int | str, str]:
    name = entry[0]
    if "/" in name:
        dir, _, file = name.partition("/")
        return 1, dir, file
    else:
        return 0, len(name), name


def hashFile(fh: t.TextIO) -> str:
    return hashlib.md5(fh.read().encode("ascii", "xmlcharrefreplace")).hexdigest()


def getDatafilePaths(basePath: str) -> t.Generator[tuple[str, str], None, None]:
    for root, dirs, files in os.walk(basePath, topdown=True):
        if "readonly" in dirs:
            dirs.remove("readonly")
        for filename in files:
            if filename == "":
                continue
            filePath = os.path.join(root, filename)
            yield filePath, os.path.relpath(filePath, basePath)


def dtNow() -> datetime:
    return datetime.now(timezone.utc)


def dtZero() -> datetime:
    return datetime(1970, 1, 1, 0, 0, 0, tzinfo=timezone.utc)


def printDt(dt: datetime) -> str:
    return dt.strftime("%Y-%m-%d %H:%M")


def formatDt(dt: datetime) -> str:
    return datetime.strftime(dt, "%Y-%m-%d %H:%M:%S.%f")


def parseDt(s: str) -> datetime | None:
    dt = trystrptime(s, "%Y-%m-%d %H:%M:%S.%f")
    if dt:
        return dt.replace(tzinfo=timezone.utc)
    return dt


def trystrptime(s: str, format: str) -> datetime | None:
    try:
        return datetime.strptime(s, format)
    except:  # pylint: disable=bare-except
        return None



# pylint: disable=attribute-defined-outside-init
from __future__ import annotations

import glob
import json
import os
import re
import sys
import time
from collections import OrderedDict, defaultdict
from datetime import datetime

from . import (
    InputSource,
    biblio,
    boilerplate,
    caniuse,
    cddl,
    conditional,
    constants,
    datablocks,
    dfns,
    doctypes,
    extensions,
    fingerprinting,
    h,
    headings,
    highlight,
    idl,
    includes,
    inlineTags,
    language,
    lint,
    markdown,
    mdn,
    metadata,
    refs,
    retrieve,
    shorthands,
    stylescript,
    t,
    wpt,
)
from . import line as l
from . import messages as m
from . import unsortedJunk as u

if t.TYPE_CHECKING:
    import widlparser


class Spec:
    def __init__(
        self,
        inputFilename: str,
        debug: bool = False,
        token: str | None = None,
        lineNumbers: bool = False,
        fileRequester: retrieve.DataFileRequester | None = None,
        testing: bool = False,
    ) -> None:
        catchArgparseBug(inputFilename)
        self.valid: bool = False
        self.lineNumbers: bool = lineNumbers
        if lineNumbers:
            # line-numbers are too hacky, so force this to be a dry run
            constants.dryRun = True
        if inputFilename is None:
            inputFilename = findImplicitInputFile()
        if inputFilename is None:  # still
            m.die(
                "No input file specified, and no *.bs or *.src.html files found in current directory.\nPlease specify an input file, or use - to pipe from STDIN.",
            )
            return
        self.inputSource: InputSource.InputSource = InputSource.inputFromName(inputFilename, chroot=constants.chroot)
        self.transitiveDependencies: set[InputSource.InputSource] = set()
        self.debug: bool = debug
        self.token: str | None = token
        self.testing: bool = testing
        self.dataFile: retrieve.DataFileRequester
        if fileRequester is None:
            self.dataFile = retrieve.defaultRequester
        else:
            self.dataFile = fileRequester

        self.lines: list[l.Line] = []
        self.valid = self.initializeState()

    def initializeState(self) -> bool:
        self.normativeRefs: dict[str, biblio.BiblioEntry] = {}
        self.informativeRefs: dict[str, biblio.BiblioEntry] = {}
        self.refs: refs.ReferenceManager = refs.ReferenceManager(fileRequester=self.dataFile, testing=self.testing)
        self.externalRefsUsed: refs.ExternalRefsManager = refs.ExternalRefsManager()

        self.md: metadata.MetadataManager
        self.mdBaseline: metadata.MetadataManager | None = metadata.MetadataManager()
        self.mdDocument: metadata.MetadataManager | None = None
        self.mdCommandLine: metadata.MetadataManager | None = metadata.MetadataManager()
        self.mdDefaults: metadata.MetadataManager | None = None
        self.mdOverridingDefaults: metadata.MetadataManager | None = None

        self.typeExpansions: dict[str, str] = {}

        self.cachedLinksFromHref: OrderedDict[str, list[t.ElementT]] = OrderedDict()
        self.cachedClassTests: dict[tuple[str, str], bool] = {}
        self.cachedNormativeEls: dict[t.ElementT, bool] = {}

        defaultMacro: t.Callable[[], str] = lambda: "???"
        self.macros: t.DefaultDict[str, str] = defaultdict(defaultMacro)

        self.widl: widlparser.Parser = idl.getParser()

        self.languages: dict[str, language.Language] = fetchLanguages(self.dataFile)
        self.doctypes: doctypes.DoctypeManager = fetchDoctypes(self.dataFile)

        self.extraJC = stylescript.JCManager()
        self.extraJC.addColors()
        self.extraJC.addMdLists()
        self.extraJC.addAutolinks()
        self.extraJC.addSelflinks()
        self.extraJC.addCounters()
        self.extraJC.addIssues()

        try:
            self.inputContent = self.inputSource.read()
            if self.inputContent.date is not None:
                self.mdBaseline.addParsedData("Date", self.inputContent.date)
        except FileNotFoundError:
            m.die(f"Couldn't find the input file at the specified location '{self.inputSource}'.")
            return False
        except OSError:
            m.die(f"Couldn't open the input file '{self.inputSource}'.")
            return False

        return True

    def initMetadata(self, inputContent: InputSource.InputContent) -> None:
        # mdDefault is already set up
        # and cli() inited the mdCommandLine

        # Get the md from the doc itself
        # TODO: currently I just leave the node in,
        #       I should do something about that
        # TODO: Pure textual hack, no way to, say, put a block
        #       in a markdown code span or an <xmp> to show off.
        _, self.mdDocument = metadata.parse(lines=inputContent.lines)

        # Combine the data so far, and compute the doctype
        # (the other md sources need the doctype in order to be found)
        self.md = metadata.join(self.mdBaseline, self.mdDocument, self.mdCommandLine)
        rawDoctype = (self.md.rawOrg, self.md.rawGroup, self.md.rawStatus)
        self.doctype = self.doctypes.getDoctype(self.md.rawOrg, self.md.rawGroup, self.md.rawStatus)

        self.mdDefaults = metadata.fromJson(
            data=retrieve.retrieveBoilerplateFile(self, "defaults"),
            source="defaults",
        )
        self.md = metadata.join(self.mdBaseline, self.mdDefaults, self.mdDocument, self.mdCommandLine)
        if rawDoctype != (self.md.rawOrg, self.md.rawGroup, self.md.rawStatus):
            # recompute doctype
            self.doctype = self.doctypes.getDoctype(self.md.rawOrg, self.md.rawGroup, self.md.rawStatus)

        # Using all of that, load up the text macros so I can sub them into the computed-metadata file.
        self.md.fillTextMacros(self.macros, doc=self)
        jsonEscapedMacros = {k: json.dumps(v)[1:-1] for k, v in self.macros.items()}

        computedMdText = h.replaceMacrosTextly(
            retrieve.retrieveBoilerplateFile(self, "computed-metadata"),
            macros=jsonEscapedMacros,
            context="? of computed-metadata.include",
        )
        self.mdOverridingDefaults = metadata.fromJson(data=computedMdText, source="computed-metadata")
        # And create the final, complete md combo
        self.md = metadata.join(
            self.mdBaseline,
            self.mdDefaults,
            self.mdOverridingDefaults,
            self.mdDocument,
            self.mdCommandLine,
        )
        # Finally, compute the "implicit" things.
        self.md.computeImplicitMetadata(doc=self)
        # And compute macros again, in case the preceding steps changed them.
        self.md.fillTextMacros(self.macros, doc=self)

        self.md.validate(doc=self)
        m.retroactivelyCheckErrorLevel()

    def earlyParse(self, inputContent: InputSource.InputContent) -> list[l.Line]:
        text = FIXMEreplaceMarkdownBlockquotes(inputContent.content)
        nodes = h.initialDocumentParse(text, h.ParseConfig.fromSpec(self))
        if self.debug:
            h.debugNodes(nodes)
        text = h.strFromNodes(nodes, withIlcc=True)
        inputContent.rawLines = [x + "\n" for x in text.split("\n")]
        return inputContent.lines

    def checkValidity(self) -> bool:
        return True

    def recordDependencies(self, *inputSources: InputSource.InputSource) -> None:
        self.transitiveDependencies.update(inputSources)

    def preprocess(self) -> Spec:
        self.transitiveDependencies.clear()
        self.assembleDocument()
        self.processDocument()
        return self

    def assembleDocument(self) -> Spec:
        self.initMetadata(self.inputContent)
        self.recordDependencies(self.inputSource)
        self.lines = self.earlyParse(self.inputContent)

        # Remove the metadata
        # FIXME: This should be done the first time I parse metadata.
        self.lines, _ = metadata.parse(lines=self.lines)
        extensions.load(self)

        # Initialize things
        self.refs.initializeRefs(doc=self, datablocks=datablocks)
        self.refs.initializeBiblio(doc=self)

        if "mixed-indents" in self.md.complainAbout:
            if self.md.indentInfo and self.md.indentInfo.char:
                checkForMixedIndents(self.lines, self.md.indentInfo)

        # Deal with further <pre> blocks, and markdown
        self.lines = datablocks.transformDataBlocks(self, self.lines)

        if "markdown-block" in self.md.markupShorthands:
            markdownFeatures: set[str] = {"headings"}
            self.lines = markdown.parse(
                self.lines,
                self.md.indent,
                opaqueElements=self.md.opaqueElements,
                blockElements=self.md.blockElements,
                features=markdownFeatures,
            )

        self.refs.setSpecData(self)

        # Convert to a single string of html now, for convenience.
        self.html = "".join(x.text for x in self.lines)
        boilerplate.addHeaderFooter(self)

        # Build the document
        self.document = h.parseDocument(self.html)
        headEl = h.find("head", self)
        bodyEl = h.find("body", self)
        assert headEl is not None
        assert bodyEl is not None
        self.head = headEl
        self.body = bodyEl
        u.correctFrontMatter(self)
        includes.processInclusions(self)
        metadata.parseDoc(self)
        return self

    def processDocument(self) -> Spec:
        # Fill in and clean up a bunch of data
        conditional.processConditionals(self)
        self.fillContainers: t.FillContainersT = u.locateFillContainers(self)
        lint.exampleIDs(self)
        wpt.processWptElements(self)

        boilerplate.addBikeshedVersion(self)
        boilerplate.addCanonicalURL(self)
        boilerplate.addFavicon(self)
        boilerplate.addSpecVersion(self)
        boilerplate.addStatusSection(self)
        boilerplate.addLogo(self)
        boilerplate.addCopyright(self)
        boilerplate.addSpecMetadataSection(self)
        boilerplate.addAbstract(self)
        boilerplate.addExpiryNotice(self)
        boilerplate.addObsoletionNotice(self)
        boilerplate.addAtRisk(self)
        u.addNoteHeaders(self)
        boilerplate.removeUnwantedBoilerplate(self)
        shorthands.run(self)
        inlineTags.processTags(self)
        u.canonicalizeShortcuts(self.body)
        u.addImplicitAlgorithms(self)
        u.fixManualDefTables(self)
        headings.processHeadings(self)
        u.checkVarHygiene(self)
        u.processIssuesAndExamples(self)
        idl.markupIDL(self)
        cddl.markupCDDL(self)
        u.inlineRemoteIssues(self)
        u.addImageSize(self)

        # Handle all the links
        u.processBiblioLinks(self)
        u.processDfns(self)
        u.processIDL(self)
        u.processCDDL(self)
        dfns.annotateDfns(self)
        u.formatArgumentdefTables(self)
        u.formatElementdefTables(self)
        u.processAutolinks(self)
        u.fixInterDocumentReferences(self)
        biblio.dedupBiblioReferences(self)
        boilerplate.addIndexSection(self)
        boilerplate.addExplicitIndexes(self)
        boilerplate.addStyles(self)
        boilerplate.addReferencesSection(self)
        boilerplate.addPropertyIndex(self)
        boilerplate.addIDLSection(self)
        boilerplate.addCDDLSection(self)
        boilerplate.addIssuesSection(self)
        boilerplate.addCustomBoilerplate(self)
        headings.processHeadings(self, "all")  # again
        boilerplate.removeUnwantedBoilerplate(self)
        boilerplate.addTOCSection(self)
        u.addSelfLinks(self)
        u.processAutolinks(self)
        boilerplate.removeUnwantedBoilerplate(self)
        # Add MDN panels after all IDs/anchors have been added
        mdn.addMdnPanels(self)
        caniuse.addCanIUsePanels(self)
        highlight.addSyntaxHighlighting(self)
        boilerplate.addBikeshedBoilerplate(self)
        boilerplate.addDarkmodeIndicators(self)
        fingerprinting.addTrackingVector(self)
        u.fixIntraDocumentReferences(self)
        u.fixInterDocumentReferences(self)
        u.verifyUsageOfAllLocalBiblios(self)
        u.removeMultipleLinks(self)
        u.forceCrossorigin(self)
        addDomintroStyles(self)
        lint.brokenLinks(self)
        lint.accidental2119(self)
        lint.missingExposed(self)
        lint.requiredIDs(self)
        lint.unusedInternalDfns(self)

        # Any final HTML cleanups
        u.cleanupHTML(self)
        if self.md.prepTR:
            # Don't try and override the W3C's icon.
            for el in h.findAll("[rel ~= 'icon']", self):
                h.removeNode(el)
            # Make sure the W3C stylesheet is after all other styles.
            for el in h.findAll("link", self):
                if el.get("href", "").startswith("https://www.w3.org/StyleSheets/TR"):
                    h.appendChild(self.head, el)
            # Ensure that all W3C links are https.
            for el in h.findAll("a", self):
                href = el.get("href", "")
                if href.startswith("http://www.w3.org") or href.startswith("http://lists.w3.org"):
                    el.set("href", "https" + href[4:])
                text = el.text or ""
                if text.startswith("http://www.w3.org") or text.startswith("http://lists.w3.org"):
                    el.text = "https" + text[4:]
            # Loaded from .include files
            extensions.BSPrepTR(self)  # type: ignore # pylint: disable=no-member

        return self

    def serialize(self) -> str | None:
        try:
            rendered = h.Serializer(self.md.opaqueElements, self.md.blockElements).serialize(self.document)
        except Exception as e:
            m.die(str(e))
            return None
        rendered = u.finalHackyCleanup(rendered)
        return rendered

    def fixMissingOutputFilename(self, outputFilename: str | None) -> str:
        if outputFilename is None:
            # More sensible defaults!
            if isinstance(self.inputSource, InputSource.TarInputSource):
                outputFilename = os.path.splitext(self.inputSource.sourceName)[0] + ".html"
            elif not isinstance(self.inputSource, InputSource.FileInputSource):
                outputFilename = "-"
            elif self.inputSource.sourceName.endswith(".bs"):
                outputFilename = self.inputSource.sourceName[0:-3] + ".html"
            elif self.inputSource.sourceName.endswith(".src.html"):
                outputFilename = self.inputSource.sourceName[0:-9] + ".html"
            else:
                outputFilename = "-"
        return outputFilename

    def finish(self, outputFilename: str | None = None, newline: str | None = None) -> None:
        # Check the errors one more time.
        m.retroactivelyCheckErrorLevel(timing="late")
        catchArgparseBug(outputFilename)
        self.printResultMessage()
        outputFilename = self.fixMissingOutputFilename(outputFilename)
        rendered = self.serialize()
        if rendered and not constants.dryRun:
            try:
                if outputFilename == "-":
                    sys.stdout.write(rendered)
                else:
                    with open(outputFilename, "w", encoding="utf-8", newline=newline) as f:
                        f.write(rendered)
            except Exception as e:
                m.die(f"Something prevented me from saving the output document to {outputFilename}:\n{e}")

    def printResultMessage(self) -> None:
        # If I reach this point, I've succeeded, but maybe with reservations.
        fatals = m.state.categoryCounts["fatal"]
        links = m.state.categoryCounts["link-error"]
        warnings = m.state.categoryCounts["warning"]
        if self.lineNumbers:
            m.warn("Because --line-numbers was used, no output was saved.")
        if fatals:
            m.success("Successfully generated, but fatal errors were suppressed")
            return
        if links:
            m.success(f"Successfully generated, with {links} linking errors")
            return
        if warnings:
            m.success("Successfully generated, with warnings")
            return

    def watch(self, outputFilename: str | None, port: int | None = None, localhost: bool = False) -> None:
        if m.wrappedOutput():
            m.die(f"Watch mode only supports console output; you've set --print={m.state.printMode}")
            return

        outputFilename = self.fixMissingOutputFilename(outputFilename)
        if self.inputSource.mtime() is None:
            m.die(f"Watch mode doesn't support {self.inputSource}")
            return
        if outputFilename == "-":
            m.die("Watch mode doesn't support streaming to STDOUT.")
            return

        if port:
            # Serve the folder on an HTTP server
            import http.server  # noqa: PLC0415
            import socketserver  # noqa: PLC0415
            import threading  # noqa: PLC0415

            class SilentServer(http.server.SimpleHTTPRequestHandler):
                def log_message(self, format: t.Any, *args: t.Any) -> None:
                    pass

            socketserver.TCPServer.allow_reuse_address = True
            server = socketserver.TCPServer(("localhost" if localhost else "", port), SilentServer)

            print(f"Serving at http://localhost:{port}/")  # noqa: T201
            thread = threading.Thread(target=server.serve_forever)
            thread.daemon = True
            thread.start()
        else:
            server = None

        mdCommandLine = self.mdCommandLine

        try:
            self.preprocess()
            self.finish(outputFilename)
            lastInputModified = {dep: dep.mtime() for dep in self.transitiveDependencies}
            printDone()
            try:
                while True:
                    # Comparing mtimes with "!=" handles when a file starts or
                    # stops existing, and it's fine to rebuild if an mtime
                    # somehow gets older.
                    if any(input.mtime() != lastModified for input, lastModified in lastInputModified.items()):
                        m.state = m.state.replace()
                        m.say("Source file modified. Rebuilding...")
                        self.initializeState()
                        self.mdCommandLine = mdCommandLine
                        self.preprocess()
                        self.finish(outputFilename)
                        lastInputModified = {dep: dep.mtime() for dep in self.transitiveDependencies}
                        printDone()
                    time.sleep(1)
            except KeyboardInterrupt:
                m.say("Exiting~")
                if server:
                    server.shutdown()
                    thread.join()
                m.printCloser()
                sys.exit(0)
        except Exception as e:
            m.die(f"Something went wrong while watching the file:\n{e}")

    def printTargets(self) -> None:
        exportMsg = "Exported terms:\n"
        for el in h.findAll("[data-export]", self):
            for term in h.linkTextsFromElement(el):
                exportMsg += "  " + term + "\n"
        m.say(exportMsg)
        unexportMsg = "Unexported terms:\n"
        for el in h.findAll("[data-noexport]", self):
            for term in h.linkTextsFromElement(el):
                unexportMsg += "  " + term + "\n"
        m.say(unexportMsg)

    def isOpaqueElement(self, el: t.ElementT) -> bool:
        if el.tag in self.md.opaqueElements:
            return True
        if el.get("data-opaque") is not None or el.get("bs-opaque") is not None:  # noqa: SIM103
            return True
        return False


def printDone() -> None:
    contents = f"Finished at {datetime.now().strftime('%H:%M:%S %b-%d-%Y')}"
    contentLen = len(contents) + 2
    if not m.state.asciiOnly:
        m.p(f"╭{'─'*contentLen}╮")
        m.p(f"│ {contents} │")
        m.p(f"╰{'─'*contentLen}╯")
        m.p("")
    else:
        m.p(f"/{'-'*contentLen}\\")
        m.p(f"| {contents} |")
        m.p(f"\\{'-'*contentLen}/")
        m.p("")


def findImplicitInputFile() -> str | None:
    """
    Find what input file the user *probably* wants to use,
    by scanning the current folder.
    In preference order:
    1. index.bs
    2. Overview.bs
    3. the first file with a .bs extension
    4. the first file with a .src.html extension
    """

    if os.path.isfile("index.bs"):
        return "index.bs"
    if os.path.isfile("Overview.bs"):
        return "Overview.bs"

    allBs = glob.glob("*.bs")
    if allBs:
        return allBs[0]

    allHtml = glob.glob("*.src.html")
    if allHtml:
        return allHtml[0]

    return None


def catchArgparseBug(string: str | None) -> bool:
    # Argparse has had a long-standing bug
    # https://bugs.python.org/issue22433
    # about spaces in the values of unknown optional arguments
    # (even when the space is in a quoted string!).
    # I can't fix this without doing a lot of work myself,
    # but I *can* discover when it has been tripped,
    # as the input or output filename will look like
    # a command-line flag, very unlikely on its own.

    if isinstance(string, str) and string.startswith("--") and "=" in string:
        m.die(
            "You're hitting a bug with Python's argparse library. Please specify both the input and output filenames manually, and move all command-line flags with spaces in their values to after those arguments.\nSee <https://speced.github.io/bikeshed/#md-issues> for details.",
        )
        return False
    return True


def fetchLanguages(dataFile: retrieve.DataFileRequester) -> dict[str, language.Language]:
    return {
        k: language.Language(v["name"], v["native-name"])
        for k, v in json.loads(dataFile.fetch("languages.json", str=True)).items()
    }


def fetchDoctypes(dataFile: retrieve.DataFileRequester) -> doctypes.DoctypeManager:
    return doctypes.DoctypeManager.fromKdlStr(dataFile.fetch("boilerplate", "doctypes.kdl", str=True))


def addDomintroStyles(doc: Spec) -> None:
    # Adds common WHATWG styles for domintro blocks.

    if h.find(".domintro", doc) is None:
        return

    doc.extraJC.addDomintro()


def checkForMixedIndents(lines: t.Sequence[l.Line], info: metadata.IndentInfo) -> None:
    badIndentChar = " " if info.char == "\t" else "\t"
    for line in lines:
        if not line.text:
            continue
        if line.text.startswith(badIndentChar):
            if info.char == " ":
                m.lint("Line starts with tabs, but document is space-indented", lineNum=line.i)
            else:
                m.lint("Line starts with spaces, but document is tab-indented", lineNum=line.i)
        if re.match(r"(\t+ +\t)|( +\t)", line.text):
            m.lint(f"Line {line.i}'s indent contains tabs after spaces.")


def FIXMEreplaceMarkdownBlockquotes(text: str) -> str:
    # Temporary hack to make the early HTML pass not be broken
    # by Markdown blockquotes.
    # * finds sequences of lines that share a ws + angle bracket prefix
    # * replaces first such > with a blockquote-open PUA
    # * replaces subsequent > with a space
    # * adds a blockquote-close PUA to the end of last line
    # * the HTML parser recognizes those PUAs and emits the correct start/end tags.

    lines = text.split("\n")
    i = 0
    while True:
        if i >= len(lines):
            break
        match = re.match(r"\s*>\s?", lines[i])
        if not match:
            i += 1
            continue
        if i + 1 < len(lines) and re.match(r"\s*>\s?", lines[i + 1]):
            lines[i] = constants.bqStart + lines[i][len(match[0]) :]
            i += 1
            while i < len(lines) and re.match(r"\s*>\s?", lines[i]):
                match = re.match(r"\s*>\s?", lines[i])
                assert match is not None
                lines[i] = lines[i][len(match[0]) :]
                i += 1
            lines[i - 1] += constants.bqEnd
        else:
            i += 1
    return "\n".join(lines)



from __future__ import annotations

import re

from .. import config, h, t


def transformMarkdownIB(doc: t.SpecT) -> None:
    # Do the remaining textual replacements

    addedNodes = []

    if "markdown-inline" not in doc.md.markupShorthands:
        return

    def transformElement(parentEl: t.ElementT) -> None:
        processContents = h.isElement(parentEl) and not doc.isOpaqueElement(parentEl)
        if not processContents:
            return
        children = h.childNodes(parentEl, clear=True)
        newChildren = []
        for el in children:
            if isinstance(el, str):
                newChildren.extend(transformText(el))
            elif h.isElement(el):
                transformElement(el)
                newChildren.append(el)
        h.appendChild(parentEl, *newChildren, allowEmpty=True)

    def transformText(text: str) -> list[t.NodeT]:
        nodes: list[t.NodeT] = [text]
        if "markdown-inline" in doc.md.markupShorthands:
            nodes = config.processTextNodes(nodes, strongRe, strongReplacer)
            nodes = config.processTextNodes(nodes, emRe, emReplacer)
            nodes = config.processTextNodes(nodes, escapedRe, escapedReplacer)
        for node in nodes:
            if h.isElement(node):
                addedNodes.append(node)
        return nodes

    transformElement(doc.document.getroot())
    for node in addedNodes:
        if h.isElement(node):
            addLineNumber(node)


def transformProductionGrammars(doc: t.SpecT) -> None:
    # Link up the various grammar symbols in CSS grammars to their definitions.
    if "css" not in doc.md.markupShorthands:
        return

    hashMultRe = re.compile(r"#{\s*\d+(\s*,(\s*\d+)?)?\s*}")

    def hashMultReplacer(match: re.Match) -> t.ElementT:
        return h.E.a({"data-link-type": "grammar", "data-lt": "#", "for": ""}, match.group(0))

    multRe = re.compile(r"{\s*\d+\s*}")

    def multReplacer(match: re.Match) -> t.ElementT:
        return h.E.a({"data-link-type": "grammar", "data-lt": "{A}", "for": ""}, match.group(0))

    multRangeRe = re.compile(r"{\s*\d+\s*,(\s*\d+)?\s*}")

    def multRangeReplacer(match: re.Match) -> t.ElementT:
        return h.E.a({"data-link-type": "grammar", "data-lt": "{A,B}", "for": ""}, match.group(0))

    simpleRe = re.compile(r"(\?|!|#|\*|\+|\|\||\||&amp;&amp;|&&|,)(?!')")
    # Note the negative-lookahead, to avoid matching delim tokens.

    def simpleReplacer(match: re.Match) -> t.ElementT:
        return h.E.a(
            {"data-link-type": "grammar", "data-lt": match.group(0), "for": ""},
            match.group(0),
        )

    addedNodes = []

    def transformElement(parentEl: t.ElementT) -> None:
        children = h.childNodes(parentEl, clear=True)
        newChildren: list[t.NodesT] = []
        for el in children:
            if isinstance(el, str):
                newChildren.extend(transformText(el))
            elif h.isElement(el):
                if el.tag != "a":
                    # Transforms all add links, which aren't allowed in <a>...
                    transformElement(el)
                newChildren.append(el)
        h.appendChild(parentEl, *newChildren, allowEmpty=True)

    def transformText(text: str) -> t.NodesT:
        nodes: list[t.NodeT] = [text]
        nodes = config.processTextNodes(nodes, hashMultRe, hashMultReplacer)
        nodes = config.processTextNodes(nodes, multRe, multReplacer)
        nodes = config.processTextNodes(nodes, multRangeRe, multRangeReplacer)
        nodes = config.processTextNodes(nodes, simpleRe, simpleReplacer)
        for node in nodes:
            if h.isElement(node):
                addedNodes.append(node)
        return nodes

    for el in h.findAll(".prod", doc):
        transformElement(el)

    for node in addedNodes:
        if h.isElement(node):
            addLineNumber(node)


strongRe = re.compile(
    r"""
                    # **, not escaped, followed by non-space
                    (?<!\\)\*\*(?!\s)
                    # Escaped **, or not a ** at all
                    ((?:[^*]|\\\*\*|\s\*\*|\*[^*])+)
                    # **, not escaped, preceded by non-space
                    (?<!\s|\\)\*\*
                    """,
    re.X,
)


def strongReplacer(match: re.Match) -> t.NodeT:
    text = t.cast(str, match.group(1)).replace("\\**", "**")
    return h.E.strong({"bs-autolink-syntax": match.group(0)}, text)


emRe = re.compile(
    r"""
                    # *, not escaped, followed by non-space
                    (?<!\\)\*(?!\s)
                    # Escaped *, or not a * at all
                    ((?:[^*]|\\\*|\s\*)+)
                    # *, not escaped, preceded by non-space
                    (?<!\s|\\)\*
                    """,
    re.X,
)


def emReplacer(match: re.Match) -> t.NodeT:
    text = t.cast(str, match.group(1)).replace("\\*", "*")
    return h.E.em({"bs-autolink-syntax": match.group(0)}, text)


escapedRe = re.compile(r"\\\*")


def escapedReplacer(match: re.Match) -> t.NodeT:  # pylint: disable=unused-argument
    return "*"


headerRe = re.compile(
    r"""
                    (\\)?
                    \[:
                    (:?[^()<>@,;:\\"/\[\]?={}\s|]+)
                    (?:\|((?:(?!:\]).)+))?
                    :\]""",
    re.X,
)


def addLineNumber(el: t.ElementT) -> None:
    if el.get("bs-line-number"):
        return
    line = h.approximateLineNumber(el)
    if line is not None:
        el.set("bs-line-number", line)
